#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeInstance:
    def test_update(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance(start=False))
        name_edited = instance.name + "-edited"
        security_group_default = SecurityGroup._from_cs(
            exo.compute,
            exo.compute.cs.listSecurityGroups(
                securitygroupname="default", fetch_list=True
            )[0],
        )

        instance.update(name=name_edited, security_groups=[security_group_default])

        [res] = exo.compute.cs.listVirtualMachines(id=instance.id, fetch_list=True)
        assert res["name"] == name_edited
        assert instance.name == name_edited

        [res] = exo.compute.cs.listSecurityGroups(
            virtualmachineid=instance.id, fetch_list=True
        )
        assert res["id"] == security_group_default.id

    def test_scale(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance(start=False))
        instance_type_small = InstanceType._from_cs(
            exo.compute.cs.listServiceOfferings(name="small", fetch_list=True)[0]
        )

        instance.scale(type=instance_type_small)

        [res] = exo.compute.cs.listVirtualMachines(id=instance.id, fetch_list=True)
        assert res["serviceofferingid"] == instance_type_small.id

    def test_start(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance(start=False))

        instance.start()

        [res] = exo.compute.cs.listVirtualMachines(id=instance.id, fetch_list=True)
        assert res["state"].lower() in {"starting", "running"}

    def test_stop(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        instance.stop()

        [res] = exo.compute.cs.listVirtualMachines(id=instance.id, fetch_list=True)
        assert res["state"].lower() in ["stopping", "stopped"]

    def test_reboot(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        instance.reboot()

        [res] = exo.compute.cs.listVirtualMachines(id=instance.id, fetch_list=True)
        assert res["state"].lower() in ["starting", "running"]

    def test_resize_volume(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance(start=False))

        instance.resize_volume(size=20)

        res = exo.compute.cs.listVolumes(virtualmachineid=instance.id, fetch_list=True)
        assert res[0]["size"] == 21474836480  # 20 GB
        assert instance.volume_size == 21474836480

    def test_snapshot_volume(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        snapshot = instance.snapshot_volume()
        assert snapshot.id != ""
        assert snapshot.date != ""
        assert snapshot.size > 0

    def test_attach_elastic_ip(self, exo, eip, instance):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(exo.compute, instance())

        instance.attach_elastic_ip(elastic_ip=elastic_ip)

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        for nic in res:
            if not nic["isdefault"]:
                continue
            assert nic["secondaryip"][0]["ipaddress"] == elastic_ip.address

    def test_detach_elastic_ip(self, exo, eip, instance):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(exo.compute, instance())

        # FIXME: attach the EIP using exo.compute.cs.addIpToNic:
        # res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        # exo.compute.cs.addIpToNic(
        #    nicid=instance._default_nic(res)["id"], ipaddress=elastic_ip.address
        # )
        instance.attach_elastic_ip(elastic_ip=elastic_ip)

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        for nic in res:
            if not nic["isdefault"]:
                continue
            assert nic["secondaryip"][0]["ipaddress"] == elastic_ip.address

        instance.detach_elastic_ip(elastic_ip)

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        for nic in res:
            if not nic["isdefault"]:
                continue
            assert "secondaryip" not in nic

    def test_attach_private_network(self, exo, privnet, instance):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        instance.attach_private_network(private_network=private_network)

        [res] = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert res["networkid"] == private_network.id
        assert res["virtualmachineid"] == instance.id

        exo.compute.cs.removeNicFromVirtualMachine(
            virtualmachineid=instance.id, nicid=res["id"]
        )

    def test_detach_private_network(self, exo, privnet, instance):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.addNicToVirtualMachine(
            virtualmachineid=instance.id, networkid=private_network.id
        )

        [res] = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert res["networkid"] == private_network.id
        assert res["virtualmachineid"] == instance.id

        instance.detach_private_network(private_network=private_network)

        res = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert len(res) == 0

    def test_set_reverse_dns(self, exo, instance, test_reverse_dns):
        instance = Instance._from_cs(exo.compute, instance())

        instance.set_reverse_dns(record=test_reverse_dns)

        res = exo.compute.cs.queryReverseDnsForVirtualMachine(id=instance.id)
        assert (
            res["virtualmachine"]["nic"][0]["reversedns"][0]["domainname"]
            == test_reverse_dns
        )

    def test_unset_reverse_dns(self, exo, instance, test_reverse_dns):
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.updateReverseDnsForVirtualMachine(
            id=instance.id, domainname=test_reverse_dns
        )
        assert (
            res["virtualmachine"]["nic"][0]["reversedns"][0]["domainname"]
            == test_reverse_dns
        )

        instance.unset_reverse_dns()

        res = exo.compute.cs.queryReverseDnsForVirtualMachine(id=instance.id)
        assert len(res["virtualmachine"]["nic"][0]["reversedns"]) == 0

    def test_delete(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance(teardown=False))
        instance_name = instance.name

        instance.delete()
        assert instance.id is None

        res = exo.compute.cs.listVirtualMachines(name=instance_name, fetch_list=True)
        assert len(res) == 0

    def test_properties(self, exo, aag, sg, privnet, eip, instance, test_reverse_dns):
        anti_affinity_group = AntiAffinityGroup._from_cs(exo.compute, aag())
        security_group = SecurityGroup._from_cs(exo.compute, sg())
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(
            exo.compute,
            instance(
                anti_affinity_groups=[anti_affinity_group.id],
                security_groups=[security_group.id],
                private_networks=[private_network.id],
                teardown=False,
            ),
        )

        instance_state = instance.state
        assert instance_state == "running"

        instance_anti_affinity_groups = list(instance.anti_affinity_groups)
        assert len(instance_anti_affinity_groups) == 1
        assert instance_anti_affinity_groups[0].id == anti_affinity_group.id

        instance_security_groups = list(instance.security_groups)
        assert len(instance_security_groups) == 1
        assert instance_security_groups[0].id == security_group.id

        instance_private_networks = list(instance.private_networks)
        assert len(instance_private_networks) == 1
        assert instance_private_networks[0].id == private_network.id

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        exo.compute.cs.addIpToNic(
            nicid=instance._default_nic(res)["id"], ipaddress=elastic_ip.address
        )
        instance_elastic_ips = list(instance.elastic_ips)
        assert len(instance_elastic_ips) == 1
        assert instance_elastic_ips[0].id == elastic_ip.id

        exo.compute.cs.updateReverseDnsForVirtualMachine(
            id=instance.id, domainname=test_reverse_dns
        )
        assert instance.reverse_dns == test_reverse_dns

        res = exo.compute.cs.createSnapshot(volumeid=instance.volume_id)
        instance_volume_snapshots = list(instance.volume_snapshots)
        assert len(instance_volume_snapshots) == 1
        assert instance_volume_snapshots[0].id == res["snapshot"]["id"]

        # We have to delete the fixture instance here because of a race condition with
        # the sg fixture teardown than will fail because it can't delete itself while
        # still used by an instance ¯\_(ツ)_/¯
        exo.compute.cs.destroyVirtualMachine(id=instance.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from cs import CloudStackApiException
from exoscale.api.dns import *


class TestDNSDomainRecord:
    def test_update(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain())
        domain_record_name = "test-python-exoscale"
        domain_record_name_edited = "test-python-exoscale-edited"
        domain_record_type = "MX"
        domain_record_content = "mx1.example.net"
        domain_record_content_edited = "mx2.example.net"
        domain_record_priority = 10
        domain_record_priority_edited = 20
        domain_record_ttl = 1042
        domain_record_ttl_edited = 1043

        exo.dns.cs.createDnsDomainRecord(
            name=domain.name,
            record_name=domain_record_name,
            record_type=domain_record_type,
            content=domain_record_content,
            ttl=domain_record_ttl,
            prio=domain_record_priority,
        )

        res = exo.dns.cs.listDnsDomainRecords(id=domain.id, fetch_list=True)
        for r in res:
            if r["name"] == "":
                continue

            domain_record = DomainRecord._from_cs(exo.dns, r, domain)

            domain_record.update(
                name=domain_record_name_edited,
                content=domain_record_content_edited,
                priority=domain_record_priority_edited,
                ttl=domain_record_ttl_edited,
            )

        res = exo.dns.cs.listDnsDomainRecords(id=domain.id, fetch_list=True)
        for r in res:
            if r["name"] == "":
                continue
            assert r["name"] == domain_record_name_edited
            assert r["content"] == domain_record_content_edited
            assert r["priority"] == domain_record_priority_edited
            assert r["ttl"] == domain_record_ttl_edited

    def test_delete(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain())
        domain_record_name = "test-python-exoscale"
        domain_record_type = "MX"
        domain_record_content = "mx1.example.net"
        domain_record_priority = 10
        domain_record_ttl = 1042

        exo.dns.cs.createDnsDomainRecord(
            name=domain.name,
            record_name=domain_record_name,
            record_type=domain_record_type,
            content=domain_record_content,
            ttl=domain_record_ttl,
            prio=domain_record_priority,
        )

        res = exo.dns.cs.listDnsDomainRecords(id=domain.id, fetch_list=True)
        for r in res:
            if r["name"] == "":
                continue

            domain_record = DomainRecord._from_cs(exo.dns, r, domain)

            domain_record.delete()

        res = exo.dns.cs.listDnsDomainRecords(id=domain.id, fetch_list=True)
        assert len(list(r for r in res if r["name"] != "")) == 0



# -*- coding: utf-8 -*-

"""
This submodule represents the Exoscale Compute API.
"""

import json
import sys
import time
from base64 import b64encode
from datetime import datetime

import attr
import requests
from cs import CloudStack, CloudStackApiException
from exoscale_auth import ExoscaleV2Auth

from . import API, APIException, RequestError, Resource, ResourceNotFoundError, polling


@attr.s
class AntiAffinityGroup(Resource):
    """
    An Anti-Affinity Group.

    Attributes:
        id (str): the Anti-Affinity Group unique identifier
        name (str): the Anti-Affinity Group display name
        description (str): the Anti-Affinity Group description
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    description = attr.ib(default="", repr=False)

    @classmethod
    def _from_cs(cls, compute, res):
        return cls(
            compute,
            res,
            id=res["id"],
            name=res["name"],
            description=res.get("description", ""),
        )

    def delete(self):
        """
        Delete the Anti-Affinity Group.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteAffinityGroup(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.id = None
        self.name = None
        self.description = None


@attr.s
class ElasticIP(Resource):
    """
    An Elastic IP.

    Attributes:
        id (str): the Elastic IP unique identifier
        zone (Zone): the zone in which the Elastic IP is located
        address (str): the Elastic IP address
        description (str): the Elastic IP description
        healthcheck_mode (str): the healthcheck probing mode (tcp|http|https)
        healthcheck_port (int): the healthcheck service port to probe
        healthcheck_path (str): the healthcheck probe HTTP request path (must be
            specified in http(s) mode)
        healthcheck_interval (int): the healthcheck probing interval in seconds
        healthcheck_timeout (int): the time in seconds before considering a healthcheck
            probing failed
        healthcheck_strikes_ok (int): the number of successful healthcheck probes before
            considering the target healthy
        healthcheck_strikes_fail (int): the number of unsuccessful healthcheck probes
            before considering the target unhealthy
        healthcheck_tls_sni (str): the TLS SNI domain to present for HTTPS healthchecks
        healthcheck_tls_skip_verify (bool): whether to skip TLS certificate validation
            for HTTPS healthchecks
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    zone = attr.ib(repr=False)
    address = attr.ib()
    description = attr.ib(default=None, repr=False)
    healthcheck_mode = attr.ib(default=None, repr=False)
    healthcheck_port = attr.ib(default=None, repr=False)
    healthcheck_path = attr.ib(default=None, repr=False)
    healthcheck_interval = attr.ib(default=None, repr=False)
    healthcheck_timeout = attr.ib(default=None, repr=False)
    healthcheck_strikes_ok = attr.ib(default=None, repr=False)
    healthcheck_strikes_fail = attr.ib(default=None, repr=False)
    healthcheck_tls_sni = attr.ib(default=None, repr=False)
    healthcheck_tls_skip_verify = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, zone=None):
        if zone is None:
            zone = compute.get_zone(id=res["zoneid"])

        return cls(
            compute,
            res,
            id=res["id"],
            zone=zone,
            address=res["ipaddress"],
            description=res.get("description", ""),
            healthcheck_mode=res.get("healthcheck", {}).get("mode", None),
            healthcheck_port=res.get("healthcheck", {}).get("port", None),
            healthcheck_path=res.get("healthcheck", {}).get("path", None),
            healthcheck_interval=res.get("healthcheck", {}).get("interval", None),
            healthcheck_timeout=res.get("healthcheck", {}).get("timeout", None),
            healthcheck_strikes_ok=res.get("healthcheck", {}).get("strikes-ok", None),
            healthcheck_strikes_fail=res.get("healthcheck", {}).get(
                "strikes-fail", None
            ),
            healthcheck_tls_sni=res.get("healthcheck", {}).get("tls-sni", None),
            healthcheck_tls_skip_verify=res.get("healthcheck", {}).get(
                "tls-skip-verify", None
            ),
        )

    @property
    def instances(self):
        """
        Instances the Elastic IP is attached to.

        Yields:
            Instance: the next instance the Elastic IP is attached to

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        for instance in self.compute.list_instances(zone=self.zone):
            for nic in instance.res.get("nic", []):
                if nic["isdefault"] and self.address in list(
                    a["ipaddress"] for a in nic.get("secondaryip", [])
                ):
                    yield instance

    @property
    def reverse_dns(self):
        """
        The reverse DNS currently set on the Elastic IP.

        Returns:
            str: reverse DNS record, or None if none set

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.compute.cs.queryReverseDnsForPublicIpAddress(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if res["publicipaddress"]["reversedns"]:
            return res["publicipaddress"]["reversedns"][0]["domainname"]

    def update(
        self,
        description=None,
        healthcheck_mode=None,
        healthcheck_port=None,
        healthcheck_path=None,
        healthcheck_interval=None,
        healthcheck_timeout=None,
        healthcheck_strikes_ok=None,
        healthcheck_strikes_fail=None,
        healthcheck_tls_sni=None,
        healthcheck_tls_skip_verify=None,
    ):
        """
        Update the Elastic IP properties.

        Parameters:
            description (str): the Elastic IP description
            healthcheck_mode (str): the healthcheck probing mode (must be either "tcp"
                or "http")
            healthcheck_port (int): the healthcheck service port to probe
            healthcheck_path (str): the healthcheck probe HTTP request path (must be
                specified in http mode)
            healthcheck_interval (int): the healthcheck probing interval in seconds
            healthcheck_timeout (int): the time in seconds before considering a
                healthcheck probing failed
            healthcheck_strikes_ok (int): the number of successful healthcheck probes
                before considering the target healthy
            healthcheck_strikes_fail (int): the number of unsuccessful healthcheck
                probes before considering the target unhealthy
            healthcheck_tls_sni (bool): the TLS SNI domain to present for HTTPS
                healthchecks
            healthcheck_tls_skip_verify (bool): whether to skip TLS certificate
                validation for HTTPS healthchecks

        Returns
            None
        """

        try:
            # We have to pass function arguments using **kwargs form because
            # of the hyphen in arguments names.
            self.compute.cs.updateIpAddress(
                **{
                    "id": self.id,
                    "description": description,
                    "mode": healthcheck_mode,
                    "port": healthcheck_port,
                    "path": healthcheck_path,
                    "interval": healthcheck_interval,
                    "timeout": healthcheck_timeout,
                    "strikes-ok": healthcheck_strikes_ok,
                    "strikes-fail": healthcheck_strikes_fail,
                    "tls-sni": healthcheck_tls_sni,
                    "tls-skip-verify": healthcheck_tls_skip_verify,
                }
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.description = description if description else self.description
        self.healthcheck_mode = (
            healthcheck_mode if healthcheck_mode else self.healthcheck_mode
        )
        self.healthcheck_port = (
            healthcheck_port if healthcheck_port else self.healthcheck_port
        )
        self.healthcheck_path = (
            healthcheck_path if healthcheck_path else self.healthcheck_path
        )
        self.healthcheck_interval = (
            healthcheck_interval if healthcheck_interval else self.healthcheck_interval
        )
        self.healthcheck_timeout = (
            healthcheck_timeout if healthcheck_timeout else self.healthcheck_timeout
        )
        self.healthcheck_strikes_ok = (
            healthcheck_strikes_ok
            if healthcheck_strikes_ok
            else self.healthcheck_strikes_ok
        )
        self.healthcheck_strikes_fail = (
            healthcheck_strikes_fail
            if healthcheck_strikes_fail
            else self.healthcheck_strikes_fail
        )
        self.healthcheck_tls_sni = (
            healthcheck_tls_sni if healthcheck_tls_sni else self.healthcheck_tls_sni
        )
        self.healthcheck_tls_skip_verify = (
            healthcheck_tls_skip_verify
            if healthcheck_tls_skip_verify
            else self.healthcheck_tls_skip_verify
        )

    def attach_instance(self, instance):
        """
        Attach the Elastic IP to a Compute instance.

        Parameters:
            instance (Instance): the instance to attach the Elastic IP to

        Returns:
            None
        """

        instance.attach_elastic_ip(elastic_ip=self)

    def detach_instance(self, instance):
        """
        Detach the Elastic IP from a Compute instance it is attached to.

        Parameters:
            instance (Instance): the instance to detach the Elastic IP from

        Returns:
            None
        """

        instance.detach_elastic_ip(elastic_ip=self)

    def set_reverse_dns(self, record):
        """
        Set the Elastic IP address reverse DNS record.

        Parameters:
            record (str): the reverse DNS record to set

        Returns:
            None
        """

        try:
            self.compute.cs.updateReverseDnsForPublicIpAddress(
                id=self.id, domainname=record
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def unset_reverse_dns(self):
        """
        Unset the Elastic IP address reverse DNS record.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteReverseDnsFromPublicIpAddress(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def delete(self, detach_instances=False):
        """
        Delete the Elastic IP.

        Parameters:
            detach_instances (bool): a flag indicating whether to detach the Elastic
                IP from the instances before deleting it

        Returns:
            None
        """

        try:
            if detach_instances:
                for instance in self.instances:
                    self.detach_instance(instance)

            self.compute.cs.disassociateIpAddress(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        # Reset all attributes
        for k, v in self.__dict__.items():
            setattr(self, k, None)


@attr.s
class Instance(Resource):
    """
    A Compute instance.

    Attributes:
        id (str): the instance unique identifier
        name (str): the instance hostname/display name
        creation_date (datetime.datetime): the instance creation date
        zone (Zone): the zone in which the instance is located
        type (InstanceType): the instance type
        template (InstanceTemplate): the instance template
        volume_size (int): the instance storage volume capacity in bytes
        ipv4_address (str): the instance public network interface IP address
        ipv6_address (str): the instance public network interface IPv6 address,
            or None if IPv6 is not enabled
        ssh_key (SSHKey): the SSH key installed in the instance user account,
        or None if no SSH key specified during instance creation
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    creation_date = attr.ib(repr=False)
    zone = attr.ib(repr=False)
    type = attr.ib(repr=False)
    template = attr.ib(repr=False)
    volume_id = attr.ib(repr=False)
    volume_size = attr.ib(repr=False)
    ipv4_address = attr.ib(repr=False)
    ipv6_address = attr.ib(default=None, repr=False)
    ssh_key = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, zone=None):
        if zone is None:
            zone = compute.get_zone(id=res["zoneid"])

        try:
            _list = compute.cs.listVolumes(virtualmachineid=res["id"], fetch_list=True)
            volume_res = _list[0]
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return cls(
            compute,
            res,
            id=res["id"],
            name=res["displayname"],
            creation_date=datetime.strptime(res["created"], "%Y-%m-%dT%H:%M:%S%z"),
            zone=zone,
            type=compute.get_instance_type(id=res["serviceofferingid"]),
            template=compute.get_instance_template(zone, id=res["templateid"]),
            volume_id=volume_res["id"],
            volume_size=volume_res["size"],
            ipv4_address=next(i for i in res["nic"] if i["isdefault"]).get(
                "ipaddress", None
            ),
            ipv6_address=next(i for i in res["nic"] if i["isdefault"]).get(
                "ip6address", None
            ),
            ssh_key=compute.get_ssh_key(name=res["keypair"])
            if "keypair" in res.keys()
            else None,
        )

    @property
    def anti_affinity_groups(self):
        """
        Anti-Affinity Groups the instance is member of.

        Yields:
            AntiAffinityGroup: the next Anti-Affinity Group the instance is member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            _list = self.compute.cs.listAffinityGroups(
                virtualmachineid=self.id, fetch_list=True
            )
            for i in _list:
                yield AntiAffinityGroup._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def elastic_ips(self):
        """
        Elastic IPs attached to the instance.

        Yields:
            ElasticIP: the next Elastic IP attached to the instance

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            # FIXME: use `iselastic=True` filter
            _list = self.compute.cs.listNics(virtualmachineid=self.id, fetch_list=True)
            default_nic = self._default_nic(_list)
            for a in default_nic.get("secondaryip", []):
                yield self.compute.get_elastic_ip(
                    zone=self.zone, address=a["ipaddress"]
                )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def private_networks(self):
        """
        Private Networks the instance is member of.

        Yields:
            PrivateNetwork: the next Private Network the instance is member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            _list = self.compute.cs.listNics(virtualmachineid=self.id, fetch_list=True)
            for nic in _list:
                if nic["isdefault"]:
                    continue
                yield self.compute.get_private_network(
                    zone=self.zone, id=nic["networkid"]
                )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def reverse_dns(self):
        """
        The reverse DNS currently set on the public network interface IP address.

        Returns:
            str: reverse DNS record, or None if none set

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.compute.cs.queryReverseDnsForVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        reverse_dns = self._default_nic(res["virtualmachine"]["nic"])["reversedns"]
        if reverse_dns:
            return reverse_dns[0]["domainname"]

    @property
    def security_groups(self):
        """
        Security Groups the instance is member of.

        Yields:
            SecurityGroup: the next Security Group the instance is member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            _list = self.compute.cs.listSecurityGroups(
                virtualmachineid=self.id, fetch_list=True
            )
            for i in _list:
                yield SecurityGroup._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def volume_snapshots(self):
        """
        Snapshots of the instance storage volume.

        Yields:
            InstanceVolumeSnapshot: the next instance storage volume snapshot

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            _list = self.compute.cs.listSnapshots(
                volumeid=self.volume_id, fetch_list=True
            )
            for i in _list:
                yield InstanceVolumeSnapshot._from_cs(self.compute, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def state(self):
        """
        State of the instance.

        Returns:
            str: the current instance state

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.listVirtualMachines(id=self.id, fetch_list=True)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return res["state"].lower()

    @property
    def instance_pool(self):
        """
        Instance Pool the instance is a member of.

        Returns:
            InstancePool: the Instance Pool the instance is member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        if self.res.get("manager") == "instancepool":
            return self.compute.get_instance_pool(self.zone, self.res["managerid"])

    def update(self, name=None, security_groups=None, user_data=None):
        """
        Update the instance properties.

        Parameters:
            name (str): an instance hostname/display name
            security_groups ([SecurityGroup]): a list of Security Groups the instance
                is member of
            user_data (str): a cloud-init user data configuration

        Returns:
            None
        """

        try:
            self.compute.cs.updateVirtualMachine(
                id=self.id,
                name=name if name is not None else None,
                displayname=name if name is not None else None,
                securitygroupids=list(i.id for i in security_groups)
                if security_groups
                else None,
                userdata=b64encode(bytes(user_data, encoding="utf-8"))
                if user_data
                else None,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if name is not None:
            self.name = name

    def scale(self, type):
        """
        Change the instance type.

        Parameters:
            type (InstanceType): the type to scale the instance to

        Returns:
            None
        """

        try:
            self.compute.cs.scaleVirtualMachine(id=self.id, serviceofferingid=type.id)
            self.type = type
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def start(self):
        """
        Start a stopped instance.

        Returns:
            None
        """

        try:
            self.compute.cs.startVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def stop(self):
        """
        Stop a running instance.

        Returns:
            None
        """

        try:
            self.compute.cs.stopVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def reboot(self):
        """
        Reboot a running instance.

        Returns:
            None
        """

        try:
            self.compute.cs.rebootVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def resize_volume(self, size):
        """
        Resize the instance storage volume.

        Parameters:
            size (int): new instance storage volume size in GB (must be greater than
                current size)

        Returns:
            None
        """

        try:
            res = self.compute.cs.resizeVolume(id=self.volume_id, size=size)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.volume_size = res["volume"]["size"]

    def snapshot_volume(self):
        """
        Take a snapshot of the instance storage volume.

        Returns:
            InstanceVolumeSnapshot: the instance storage volume snapshot taken
        """

        try:
            res = self.compute.cs.createSnapshot(volumeid=self.volume_id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return InstanceVolumeSnapshot._from_cs(self.compute, res["snapshot"])

    def attach_elastic_ip(self, elastic_ip):
        """
        Attach an Elastic IP to the instance.

        Parameters:
            elastic_ip (ElasticIP): the Elastic IP to attach

        Returns:
            None
        """

        try:
            _list = self.compute.cs.listNics(virtualmachineid=self.id, fetch_list=True)
            default_nic = self._default_nic(_list)
            self.compute.cs.addIpToNic(
                nicid=default_nic["id"], ipaddress=elastic_ip.address
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def detach_elastic_ip(self, elastic_ip):
        """
        Detach an Elastic IP from the instance.

        Parameters:
            elastic_ip (ElasticIP): the Elastic IP to detach

        Returns:
            None
        """

        try:
            _list = self.compute.cs.listNics(virtualmachineid=self.id, fetch_list=True)
            default_nic = self._default_nic(_list)
            for a in default_nic.get("secondaryip", []):
                if a["ipaddress"] == elastic_ip.address:
                    self.compute.cs.removeIpFromNic(id=a["id"])
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def attach_private_network(self, private_network):
        """
        Attach the instance to a Private Network.

        Parameters:
            private_network (PrivateNetwork): the Private Network to attach to

        Returns:
            None
        """

        try:
            self.compute.cs.addNicToVirtualMachine(
                virtualmachineid=self.id, networkid=private_network.id
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def detach_private_network(self, private_network):
        """
        Detach the instance from a Private Network.

        Parameters:
            private_network (PrivateNetwork): the Private Network to detach from

        Returns:
            None
        """

        try:
            [res] = self.compute.cs.listNics(
                virtualmachineid=self.id, networkid=private_network.id, fetch_list=True
            )

            self.compute.cs.removeNicFromVirtualMachine(
                virtualmachineid=self.id, nicid=res["id"]
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def set_reverse_dns(self, record):
        """
        Set the public network interface IP address reverse DNS record.

        Parameters:
            record (str): the reverse DNS record to set

        Returns:
            None
        """

        try:
            self.compute.cs.updateReverseDnsForVirtualMachine(
                id=self.id, domainname=record
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def unset_reverse_dns(self):
        """
        Unset the public network interface IP address reverse DNS record.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteReverseDnsFromVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def delete(self):
        """
        Delete the instance.

        Returns:
            None
        """

        try:
            self.compute.cs.destroyVirtualMachine(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.id = None
        self.name = None
        self.creation_date = None
        self.zone = None
        self.type = None
        self.template = None
        self.ipv4_address = None
        self.ipv6_address = None
        self.ssh_key = None

    def _default_nic(self, nics):
        for nic in nics:
            if nic["isdefault"]:
                return nic


@attr.s
class InstanceTemplate(Resource):
    """
    A Compute instance template.

    Attributes:
        id (str): the template unique identifier
        name (str): the template name
        description (str): the template description
        zone (Zone): the zone in which the template is located
        date (datetime.datetime): the template creation date
        size (int): the template disk size
        ssh_key_enabled (bool): a flag indicating whether the SSH key deployment is
            enabled
        password_reset_enabled (bool): a flag indicating whether the user password can
            be reset

    References:
        * `Python datetime module`_

        .. _Python datetime module: https://docs.python.org/3/library/datetime.html
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    description = attr.ib(repr=False)
    zone = attr.ib(repr=False)
    date = attr.ib(repr=False)
    size = attr.ib(repr=False)
    ssh_key_enabled = attr.ib(default=True, repr=False)
    password_reset_enabled = attr.ib(default=True, repr=False)
    username = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, zone=None):
        if zone is None:
            zone = compute.get_zone(id=res["zoneid"])

        return cls(
            compute,
            res,
            id=res["id"],
            name=res["name"],
            description=res.get("displaytext", ""),
            zone=zone,
            date=datetime.strptime(res["created"], "%Y-%m-%dT%H:%M:%S%z"),
            size=res["size"],
            username=res.get("details", {}).get("username", None),
            ssh_key_enabled=res["sshkeyenabled"],
            password_reset_enabled=res["passwordenabled"],
        )

    @classmethod
    def _register(
        cls,
        compute,
        name,
        url,
        checksum,
        zone,
        description=None,
        username=None,
        disable_ssh_key=False,
        disable_password_reset=False,
    ):
        templateDetails = {}
        if username is not None:
            templateDetails["username"] = username

        try:
            res = compute.cs.registerCustomTemplate(
                name=name,
                displaytext=description,
                zoneId=zone.id,
                url=url,
                checksum=checksum,
                details=templateDetails,
                sshkeyenabled=False if disable_ssh_key else True,
                passwordenabled=False if disable_password_reset else True,
                fetch_result=True,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return cls._from_cs(compute, res["template"][0])

    def delete(self):
        """
        Delete the instance template.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteTemplate(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.name = None
        self.description = None
        self.zone = None
        self.date = None
        self.ssh_key_enabled = None
        self.password_reset_enabled = None
        self.username = None


@attr.s
class InstanceVolumeSnapshot(Resource):
    """
    A Compute instance storage volume snapshot.

    Attributes:
        id (str): the instance storage volume snapshot unique identifier
        date (datetime.datetime): the instance storage volume snapshot creation date
        size (int): the instance storage volume snapshot size in bytes

    References:
        * `Python datetime module`_

        .. _Python datetime module: https://docs.python.org/3/library/datetime.html
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    date = attr.ib(repr=False)
    size = attr.ib(repr=False)

    @classmethod
    def _from_cs(cls, compute, res):
        return cls(
            compute,
            res,
            id=res["id"],
            date=datetime.strptime(res["created"], "%Y-%m-%dT%H:%M:%S%z"),
            size=res["size"],
        )

    @property
    def state(self):
        """
        State of the instance storage volume snapshot.

        Returns:
            str: the current instance storage volume snapshot state.

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.listSnapshots(id=self.id, fetch_list=True)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return res["state"].lower()

    def revert(self):
        """
        Revert the storage volume snapshot.

        Returns:
            None
        """

        try:
            res = self.compute.cs.revertSnapshot(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if not res["success"]:
            raise APIException(reason=res["displaytext"])

    def delete(self):
        """
        Delete the storage volume snapshot.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteSnapshot(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.id = None
        self.date = None
        self.size = None

    def export(self):
        """
        Exports the storage volume snapshot.

        Returns:
            dict: the exported snapshot file properties (URL, checksum...)
        """

        try:
            res = self.compute.cs.exportSnapshot(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return res


@attr.s
class InstanceType(Resource):
    """
    A Compute instance type.

    Attributes:
        id (str): the instance type unique identifier
        name (str): the instance type name
        cpu (int): the number of vCPU allocated
        memory (int): the amount of RAM allocated in MB
    """

    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    cpu = attr.ib(repr=False)
    memory = attr.ib(repr=False)

    @classmethod
    def _from_cs(cls, res):
        return cls(
            res,
            id=res["id"],
            name=res["name"],
            cpu=res["cpunumber"],
            memory=res["memory"],
        )


@attr.s
class InstancePool(Resource):
    """
    A Compute Instance Pool.

    Attributes:
        id (str): the Instance Pool unique identifier
        name (str): the Instance Pool name
        description (str): the Instance Pool description
        zone (Zone): the zone in which the Instance Pool is located
        size (int): the number of Compute instance members the Instance Pool manages
        instance_type (InstanceType): the type of instances managed by this instance
            pool
        instance_template (InstanceTemplate): the template to be used when this instance
            pool creates new instances.
        instance_user_data (InstanceTemplate): The base64-encoded instances user
            data, when the Instance Pool creates new instances
        instance_volume_size (int): the storage volume capacity in bytes to set when
            this Instance Pool creates new instances.
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    zone = attr.ib(repr=False)
    size = attr.ib(repr=False)
    instance_type = attr.ib(repr=False)
    instance_template = attr.ib(repr=False)
    instance_user_data = attr.ib(repr=False)
    instance_volume_size = attr.ib(repr=False)
    description = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, zone=None):
        if zone is None:
            zone = compute.get_zone(id=res["zoneid"])

        return cls(
            compute,
            res,
            id=res["id"],
            name=res["name"],
            description=res.get("description"),
            zone=zone,
            size=res["size"],
            instance_type=compute.get_instance_type(id=res["serviceofferingid"]),
            instance_template=compute.get_instance_template(zone, id=res["templateid"]),
            instance_user_data=res.get("userdata"),
            instance_volume_size=res["rootdisksize"],
        )

    @property
    def instances(self):
        """
        Pool instance members.

        Yields:
            Instance: the next instance managed by this pool

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.getInstancePool(
                id=self.id, zoneid=self.zone.id, fetch_list=True
            )
            for i in res.get("virtualmachines", []):
                yield Instance._from_cs(self.compute, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def anti_affinity_groups(self):
        """
        Anti-Affinity Groups the instances are member of.

        Yields:
            AntiAffinityGroup: the next Anti-Affinity Group the instances are member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.getInstancePool(
                id=self.id, zoneid=self.zone.id, fetch_list=True
            )
            for aagid in res.get("affinitygroupids", []):
                yield self.compute.get_anti_affinity_group(id=aagid)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def security_groups(self):
        """
        Security Groups the instances are member of.

        Yields:
            SecurityGroup: the next Security Group the instances are member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.getInstancePool(
                id=self.id, zoneid=self.zone.id, fetch_list=True
            )
            for sgid in res.get("securitygroupids", []):
                yield self.compute.get_security_group(id=sgid)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def private_networks(self):
        """
        Private Networks the instances are member of.

        Yields:
            PrivateNetwork: the next Private Network the instances are member of

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.getInstancePool(
                id=self.id, zoneid=self.zone.id, fetch_list=True
            )
            for nid in res.get("networkids", []):
                yield self.compute.get_private_network(zone=self.zone, id=nid)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def state(self):
        """
        State of the Instance Pool.

        Returns:
            str: the current Instance Pool state.

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.getInstancePool(
                id=self.id, zoneid=self.zone.id, fetch_list=True
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return res["state"].lower()

    def scale(self, size):
        """
        Scale the Instance Pool up or down.

        Parameters:
            size (int): the number of Compute instance members the pool must manage

        Returns:
            None
        """

        if size <= 0:
            raise ValueError("size must be > 0")

        try:
            self.compute.cs.scaleInstancePool(
                id=self.id, zoneid=self.zone.id, size=size
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.size = size

    def update(
        self,
        name=None,
        description=None,
        instance_template=None,
        instance_volume_size=None,
        instance_user_data=None,
    ):
        """
        Update the Instance Pool properties.

        Parameters:
            name (str): an Instance Pool name
            description (str): an Instance Pool description
            instance_template (InstanceTemplate): an instance template to use for
                Compute instance members
            instance_volume_size (int): the Compute instance members storage volume
                size in GB
            instance_user_data (str): a cloud-init user data configuration to apply to
                the Compute instance members

        Returns:
            None
        """

        instance_template_id = None
        if instance_template is not None:
            instance_template_id = instance_template.id

        instance_user_data_content = None
        if instance_user_data is not None:
            instance_user_data_content = b64encode(
                bytes(instance_user_data, encoding="utf-8")
            )

        try:
            self.compute.cs.updateInstancePool(
                id=self.id,
                zoneid=self.zone.id,
                name=name,
                description=description,
                templateid=instance_template_id,
                rootdisksize=instance_volume_size,
                userdata=instance_user_data_content,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if name:
            self.name = name

        if description:
            self.description = description

        if instance_template is not None:
            self.instance_template = instance_template

        if instance_user_data_content:
            self.instance_user_data = instance_user_data_content
        if instance_volume_size is not None:
            self.instance_volume_size = instance_volume_size

    def delete(self, wait=True, max_poll=300):
        """
        Delete the Instance Pool.

        Returns:
            None
        """

        try:
            self.compute.cs.destroyInstancePool(id=self.id, zoneid=self.zone.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if wait:
            for _ in range(max_poll):
                time.sleep(1)
                try:
                    self.compute.cs.getInstancePool(id=self.id, zoneid=self.zone.id)
                except CloudStackApiException as e:
                    if e.error["errorcode"] == 404:
                        break
                    else:
                        raise APIException(e.error["errortext"], e.error)
        self.compute = None
        self.res = None
        self.id = None
        self.name = None
        self.zone = None
        self.size = None
        self.instance_type = None
        self.instance_template = None
        self.instance_user_data = None
        self.instance_volume_size = None


@attr.s
class NetworkLoadBalancerServiceHealthcheck(Resource):
    """
    A Network Load Balancer service healthcheck.

    Attributes:
        mode (str): the healthcheck probing mode (tcp|http|https)
        port (int): the healthcheck service port to probe
        uri (str): the healthcheck probe HTTP request path (must be specified in http(s)
            mode)
        interval (int): the healthcheck probing interval in seconds
        timeout (int): the time in seconds before considering a healthcheck
            probing failed
        retries (int): the number of times to retry a failed healthchecking probe before
            marking the target as failing
        tls_sni (str): the TLS SNI domain to present for HTTPS healthchecks
    """

    res = attr.ib(repr=False)
    mode = attr.ib()
    port = attr.ib()
    uri = attr.ib(default=None, repr=False)
    interval = attr.ib(default=None, repr=False)
    timeout = attr.ib(default=None, repr=False)
    retries = attr.ib(default=None, repr=False)
    tls_sni = attr.ib(default=None, repr=False)

    @classmethod
    def _from_api(cls, res):
        return cls(
            res,
            mode=res["mode"],
            port=res["port"],
            uri=res.get("uri"),
            interval=res.get("interval"),
            timeout=res.get("timeout"),
            retries=res.get("retries"),
            tls_sni=res.get("tls-sni"),
        )


@attr.s
class NetworkLoadBalancerService(Resource):
    """
    A Network Load Balancer service.

    Attributes:
        nlb (NetworkLoadBalancer): the parent Network Load Balancer instance
        id (str): the Network Load Balancer service uniquer identifier
        name (str): the Network Load Balancer service name
        description (str): a Network Load Balancer service description
        instance_pool (InstancePool): the Instance Pool to forward the Network Load
            Balancer service traffic to
        port (int): the Network Load Balancer service port
        target_port (port): the port to forward the Network Load Balancer service
            traffic to
        protocol (str): the Network Load Balancer service protocol (tcp|udp)
        strategy (str): the Network Load Balancer service dispatch strategy
            (round-robin|source-hash)
        healthcheck (NetworkLoadBalancerServiceHealthcheck): the Network Load Balancer
            service healthcheck
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    nlb = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    instance_pool = attr.ib(repr=False)
    port = attr.ib(repr=False)
    target_port = attr.ib(repr=False)
    protocol = attr.ib(repr=False)
    strategy = attr.ib(repr=False)
    healthcheck = attr.ib(repr=False)
    description = attr.ib(default="", repr=False)

    @classmethod
    def _from_api(cls, compute, res, nlb):
        return cls(
            compute,
            res,
            nlb=nlb,
            id=res["id"],
            name=res["name"],
            description=res.get("description"),
            instance_pool=compute.get_instance_pool(
                nlb.zone, res["instance-pool"]["id"]
            ),
            port=res["port"],
            target_port=res["target-port"],
            protocol=res["protocol"],
            strategy=res["strategy"],
            healthcheck=NetworkLoadBalancerServiceHealthcheck._from_api(
                res["healthcheck"]
            ),
        )

    @property
    def healthcheck_status(self):
        """
        Status of the Network Load Balancer service healthcheck.

        Returns:
            [dict]: the current Network Load Balancer service healthcheck status,
                a list of dicts containing keys "public-ip" and "status".

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.compute._v2_request(
            "GET",
            "/load-balancer/{}/service/{}".format(self.nlb.id, self.id),
            self.nlb.zone.name,
        )

        return res["healthcheck-status"]

    @property
    def state(self):
        """
        State of the Network Load Balancer service.

        Returns:
            str: the current Network Load Balancer service state.

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.compute._v2_request(
            "GET",
            "/load-balancer/{}/service/{}".format(self.nlb.id, self.id),
            self.nlb.zone.name,
        )

        return res["state"]

    def update(
        self,
        name=None,
        description=None,
        port=None,
        target_port=None,
        protocol=None,
        strategy=None,
        healthcheck_mode=None,
        healthcheck_port=None,
        healthcheck_uri=None,
        healthcheck_interval=None,
        healthcheck_timeout=None,
        healthcheck_retries=None,
        healthcheck_tls_sni=None,
    ):
        """
        Update the Network Load Balancer service properties.

        Parameters:
            name (str): the Network Load Balancer service name
            description (str): a Network Load Balancer service description
            port (int): the Network Load Balancer service port
            target_port (port): the port to forward the Network Load Balancer service
                traffic to
            protocol (str): the Network Load Balancer service protocol (tcp|udp)
            strategy (str): the Network Load Balancer service dispatch strategy
                (round-robin|source-hash)
            healtcheck_mode (str): the healthcheck probing mode (tcp|http|https)
            healtcheck_port (int): the healthcheck service port to probe
            healtcheck_uri (str): the healthcheck probe HTTP request path (must be
                specified in http(s) mode)
            healthcheck_interval (int): the healthcheck probing interval in seconds
            healthcheck_timeout (int): the time in seconds before considering a
                healthcheck probing failed
            healthcheck_retries (int): the number of times to retry a failed
                healthchecking probe before marking the target as failing
            healthcheck_tls_sni (str): the TLS SNI domain to present for HTTPS
                healthchecks
        Returns:
            None
        """

        self.compute._v2_request_async(
            "PUT",
            "/load-balancer/{}/service/{}".format(self.nlb.id, self.id),
            zone=self.nlb.zone.name,
            json={
                "name": name if name is not None else self.name,
                "description": description
                if description is not None
                else self.description,
                "port": port if port is not None else self.port,
                "target-port": target_port
                if target_port is not None
                else self.target_port,
                "protocol": protocol if protocol is not None else self.protocol,
                "strategy": strategy if strategy is not None else self.strategy,
                "healthcheck": {
                    "mode": healthcheck_mode
                    if healthcheck_mode is not None
                    else self.healthcheck.mode,
                    "port": healthcheck_port
                    if healthcheck_port is not None
                    else self.healthcheck.port,
                    "uri": healthcheck_uri
                    if healthcheck_uri is not None
                    else self.healthcheck.uri,
                    "interval": healthcheck_interval
                    if healthcheck_interval is not None
                    else self.healthcheck.interval,
                    "timeout": healthcheck_timeout
                    if healthcheck_timeout is not None
                    else self.healthcheck.timeout,
                    "retries": healthcheck_retries
                    if healthcheck_retries is not None
                    else self.healthcheck.retries,
                    "tls-sni": healthcheck_tls_sni
                    if healthcheck_tls_sni is not None
                    else self.healthcheck.tls_sni,
                },
            },
        )

        if name is not None:
            self.name = name
        if description is not None:
            self.description = description
        if port is not None:
            self.port = port
        if target_port is not None:
            self.target_port = target_port
        if protocol is not None:
            self.protocol = protocol
        if strategy is not None:
            self.strategy = strategy
        if healthcheck_mode is not None:
            self.healthcheck.mode = healthcheck_mode
        if healthcheck_port is not None:
            self.healthcheck.port = healthcheck_port
        if healthcheck_uri is not None:
            self.healthcheck.uri = healthcheck_uri
        if healthcheck_interval is not None:
            self.healthcheck.interval = healthcheck_interval
        if healthcheck_timeout is not None:
            self.healthcheck.timeout = healthcheck_timeout
        if healthcheck_retries is not None:
            self.healthcheck.retries = healthcheck_retries
        if healthcheck_tls_sni is not None:
            self.healthcheck.tls_sni = healthcheck_tls_sni

    def delete(self):
        """
        Delete the Network Load Balancer service.

        Returns:
            None
        """

        res = self.compute._v2_request_async(
            "DELETE",
            "/load-balancer/{}/service/{}".format(self.nlb.id, self.id),
            self.nlb.zone.name,
        )

        # Reset all attributes
        for k, v in self.__dict__.items():
            setattr(self, k, None)


@attr.s
class NetworkLoadBalancer(Resource):
    """
    A Network Load Balancer.

    Attributes:
        id (str): the Network Load Balancer unique identifier
        name (str): the Network Load Balancer name
        description (str): a Network Load Balancer description
        creation_date (datetime.datetime): the Network Load Balancer creation date
        ip_address (str): the Network Load Balancer public IP address
        zone (Zone): the zone in which the Network Load Balancer is located
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    creation_date = attr.ib(repr=False)
    ip_address = attr.ib(repr=False)
    zone = attr.ib(repr=False)
    description = attr.ib(default="", repr=False)

    @classmethod
    def _from_api(cls, compute, res, zone):
        return cls(
            compute,
            res,
            id=res["id"],
            zone=zone,
            name=res["name"],
            description=res["description"],
            creation_date=datetime.strptime(res["created-at"], "%Y-%m-%dT%H:%M:%SZ"),
            ip_address=res["ip"],
        )

    @property
    def services(self):
        """
        Services running on the Network Load Balancer.

        Yields:
            NetworkLoadBalancerService: the next Network Load Balancer service

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.compute._v2_request(
            "GET", "/load-balancer/" + self.id, self.zone.name
        )

        for svc in res["services"]:
            yield NetworkLoadBalancerService._from_api(self.compute, svc, self)

    @property
    def state(self):
        """
        State of the Network Load Balancer.

        Returns:
            str: the current Network Load Balancer state.

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.compute._v2_request(
            "GET", "/load-balancer/" + self.id, self.zone.name
        )

        return res["state"]

    def add_service(
        self,
        name,
        instance_pool,
        port,
        healthcheck_interval,
        protocol="tcp",
        description="",
        target_port=None,
        strategy="round-robin",
        healthcheck_mode="tcp",
        healthcheck_port=None,
        healthcheck_uri=None,
        healthcheck_timeout=None,
        healthcheck_retries=None,
        healthcheck_tls_sni=None,
    ):
        """
        Add a new service to the Network Load Balancer.

        Parameters:
            name (str): the Network Load Balancer service name
            description (str): a Network Load Balancer service description
            instance_pool (InstancePool): the Instance Pool to forward the Network Load
                Balancer service traffic to
            port (int): the Network Load Balancer service port
            target_port (port): the port to forward the Network Load Balancer service
                traffic to
            protocol (str): the Network Load Balancer service protocol (tcp|udp)
            strategy (str): the Network Load Balancer service dispatch strategy
                (round-robin|source-hash)
            healtcheck_mode (str): the healthcheck probing mode (tcp|http|https)
            healtcheck_port (int): the healthcheck service port to probe
            healtcheck_uri (str): the healthcheck probe HTTP request path (must be
                specified in http(s) mode)
            healthcheck_interval (int): the healthcheck probing interval in seconds
            healthcheck_timeout (int): the time in seconds before considering a
                healthcheck probing failed
            healthcheck_retries (int): the number of times to retry a failed
                healthchecking probe before marking the target as failing
            healthcheck_tls_sni (str): the TLS SNI domain to present for HTTPS
                healthchecks

        Returns:
            NetworkLoadBalancerService: the Network Load Balancer service added.
        """

        # The API doesn't return the NLB service created directly, so in order to return
        # a NetworkLoadBalancerService corresponding to the new service we have to
        # manually compare the list of services on the NLB instance before and after the
        # service creation, and identify the service that wasn't there before.
        # Note: in case of multiple services creation in parallel this technique is
        # subject to race condition as we could return an unrelated service. To prevent
        # this, we also compare the name of the new service to the name specified in the
        # parameters.
        services = []
        for svc in self.services:
            services.append(svc.id)

        if target_port is None:
            target_port = port
        if healthcheck_port is None:
            healthcheck_port = target_port

        res = self.compute._v2_request_async(
            "POST",
            "/load-balancer/{}/service".format(self.id),
            zone=self.zone.name,
            json={
                "name": name,
                "description": description,
                "instance-pool": {"id": instance_pool.id},
                "port": port,
                "target-port": target_port,
                "protocol": protocol,
                "strategy": strategy,
                "healthcheck": {
                    "mode": healthcheck_mode,
                    "port": healthcheck_port,
                    "uri": healthcheck_uri,
                    "interval": healthcheck_interval,
                    "timeout": healthcheck_timeout,
                    "retries": healthcheck_retries,
                    "tls-sni": healthcheck_tls_sni,
                },
            },
        )

        # Look for an unknown service: if we find one we hope it's the one we've just
        # created.
        for svc in self.services:
            if svc.id not in services and svc.name == name:
                return svc

        raise APIException("unable to retrieve the service created")

    def update(self, name=None, description=None):
        """
        Update the Network Load Balancer properties.

        Parameters:
            name (str): the Network Load Balancer name
            description (str): the Network Load Balancer description

        Returns:
            None
        """

        res = self.compute._v2_request_async(
            "PUT",
            "/load-balancer/" + self.id,
            zone=self.zone.name,
            json={"name": name, "description": description},
        )

        if name is not None:
            self.name = name
        if description is not None:
            self.description = description

    def delete(self):
        """
        Delete the Network Load Balancer.

        Returns:
            None
        """

        res = self.compute._v2_request_async(
            "DELETE", "/load-balancer/" + self.id, self.zone.name
        )

        # Reset all attributes
        for k, v in self.__dict__.items():
            setattr(self, k, None)


@attr.s
class PrivateNetwork(Resource):
    """
    A Private Network.

    Attributes:
        id (str): the Private Network unique identifier
        name (str): the Private Network name
        description (str): the Private Network description
        zone (Zone): the zone in which the Private Network is located
        start_ip (str): the start address of the managed Private Network IP range
        end_ip (str): the end address of the managed Private Network IP range
        netmask (str): the managed Private Network IP range netmask

    Note:
        The ``start_ip``, ``end_ip`` and ``netmask`` attributes are required in
        "managed" mode.
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    zone = attr.ib(repr=False)
    description = attr.ib(default="", repr=False)
    start_ip = attr.ib(default=None, repr=False)
    end_ip = attr.ib(default=None, repr=False)
    netmask = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, zone=None):
        if zone is None:
            zone = compute.get_zone(id=res["zoneid"])

        return cls(
            compute,
            res,
            id=res["id"],
            zone=zone,
            name=res["name"],
            description=res["displaytext"],
            start_ip=res.get("startip", ""),
            end_ip=res.get("endip", ""),
            netmask=res.get("netmask", ""),
        )

    @property
    def instances(self):
        """
        Private Network instance members.

        Yields:
            Instance: the next Private Network instance member

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        return self.compute.list_instances(zone=self.zone, networkid=self.id)

    def update(
        self, name=None, description=None, start_ip=None, end_ip=None, netmask=None
    ):
        """
        Update the Private Network properties.

        Parameters:
            name (str): a Private Network name
            description (str): a Private Network description
            start_ip (str): a start address of the managed Private Network IP range
            end_ip (str): an end address of the managed Private Network IP range
            netmask (str): a managed Private Network IP range netmask

        Returns:
            None
        """

        try:
            self.compute.cs.updateNetwork(
                id=self.id,
                name=name,
                displaytext=description,
                startip=start_ip,
                endip=end_ip,
                netmask=netmask,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        if name is not None:
            self.name = name
        if description is not None:
            self.description = description
        if start_ip is not None:
            self.start_ip = start_ip
        if end_ip is not None:
            self.end_ip = end_ip
        if netmask is not None:
            self.netmask = netmask

    def attach_instance(self, instance):
        """
        Attach a Compute instance to the Private Network.

        Parameters:
            instance (Instance): the instance to attach

        Returns:
            None
        """

        instance.attach_private_network(self)

    def detach_instance(self, instance):
        """
        Detach a Compute instance from the Private Network.

        Parameters:
            instances (Instance): the instance to detach

        Returns:
            None
        """

        instance.detach_private_network(self)

    def delete(self):
        """
        Delete the Private Network.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteNetwork(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.id = None
        self.name = None
        self.zone = None
        self.description = None
        self.start_ip = None
        self.end_ip = None
        self.netmask = None


@attr.s
class SecurityGroup(Resource):
    """
    A Security Group.

    Attributes:
        id (str): the Security Group unique identifier
        name (str): the Security Group name
        description (str): the Security Group description
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    description = attr.ib(default="", repr=False)

    @classmethod
    def _from_cs(cls, compute, res):
        return cls(
            compute,
            res,
            id=res["id"],
            name=res["name"],
            description=res.get("description", ""),
        )

    @property
    def ingress_rules(self):
        """
        Ingress rules of the Security Group.

        Yields:
            SecurityGroupRule: the next ingress rule of the Security Group

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.listSecurityGroups(id=self.id, fetch_list=True)
            for rule in res.get("ingressrule", []):
                yield SecurityGroupRule._from_cs(
                    type="ingress", compute=self.compute, res=rule
                )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    @property
    def egress_rules(self):
        """
        Egress rules of the Security Group.

        Yields:
            SecurityGroupRule: the next egress rule of the Security Group

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            [res] = self.compute.cs.listSecurityGroups(id=self.id, fetch_list=True)
            for rule in res.get("egressrule", []):
                yield SecurityGroupRule._from_cs(
                    type="egress", compute=self.compute, res=rule
                )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def add_rule(self, rule):
        """
        Add a rule to the Security Group.

        Parameters:
            rule (SecurityGroupRule): the Security Group rule to add

        Returns:
            None
        """

        if rule.type not in {"ingress", "egress"}:
            raise ValueError("rule type must be either ingress or egress")

        start_port, end_port = rule._parse_port()

        try:
            if rule.type == "ingress":
                self.compute.cs.authorizeSecurityGroupIngress(
                    securitygroupid=self.id,
                    description=rule.description,
                    cidrlist=rule.network_cidr,
                    startport=start_port,
                    endport=end_port,
                    icmpcode=rule.icmp_code,
                    icmptype=rule.icmp_type,
                    protocol=rule.protocol,
                    usersecuritygrouplist={"group": rule.security_group.name}
                    if rule.security_group
                    else None,
                )
            else:
                self.compute.cs.authorizeSecurityGroupEgress(
                    securitygroupid=self.id,
                    description=rule.description,
                    cidrlist=rule.network_cidr,
                    startport=start_port,
                    endport=end_port,
                    icmpcode=rule.icmp_code,
                    icmptype=rule.icmp_type,
                    protocol=rule.protocol,
                    usersecuritygrouplist={"group": rule.security_group.name}
                    if rule.security_group
                    else None,
                )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def delete(self):
        """
        Delete the Security Group.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteSecurityGroup(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.compute = None
        self.res = None
        self.id = None
        self.name = None
        self.description = None


@attr.s
class SecurityGroupRule:
    """
    A Security Group rule.

    Attributes:
        id (str): the Security Group rule unique identifier
        type (str): the Security Group rule type (ingress or egress)
        description (str): the Security Group rule description
        network_cidr (str): a source/destination network CIDR to match
            (conflicts with security_group)
        security_group (SecurityGroup): a source/destination Security Group to match
            (conflicts with network_cidr)
        port (str): the source/destination port or port range to match
        protocol (str): a network protocol to match
        icmp_code (int): an ICMP code to match
        icmp_type (int): an ICMP type to match
    """

    type = attr.ib()
    compute = attr.ib(default=None, repr=False)
    id = attr.ib(default=None)
    description = attr.ib(default=None, repr=False)
    network_cidr = attr.ib(default=None, repr=False)
    security_group = attr.ib(default=None, repr=False)
    port = attr.ib(default=None, repr=False)
    protocol = attr.ib(default="tcp", repr=False)
    icmp_code = attr.ib(default=None, repr=False)
    icmp_type = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res, type):
        port = str(res.get("startport", ""))
        port = (
            "-".join([port, str(res["endport"])])
            if res.get("startport", None) is not None and str(res["endport"]) != port
            else port
        )

        return cls(
            type=type,
            compute=compute,
            id=res["ruleid"],
            description=res.get("description", None),
            network_cidr=res.get("cidr", None),
            security_group=compute.get_security_group(name=res["securitygroupname"])
            if "securitygroupname" in res
            else None,
            port=port,
            protocol=res.get("protocol", None),
            icmp_code=res.get("icmp_code", None),
            icmp_type=res.get("icmp_type", None),
        )

    @classmethod
    def ingress(cls, **kwargs):
        """
        Returns an ingress-type SecurityGroupRule object.

        Returns:
            SecurityGroupRule: an ingress-type Security Group rule

        See Also:
            See `SecurityGroupRule <#exoscale.api.compute.SecurityGroupRule>`_ class
            documentation for parameters.
        """

        return cls(type="ingress", **kwargs)

    @classmethod
    def egress(cls, **kwargs):
        """
        Returns an egress-type SecurityGroupRule object.

        Returns:
            SecurityGroupRule: an egress-type Security Group rule

        See Also:
            See `SecurityGroupRule <#exoscale.api.compute.SecurityGroupRule>`_ class
            documentation for parameters.
        """

        return cls(type="egress", **kwargs)

    def delete(self):
        """
        Delete the Security Group rule.

        Returns:
            None
        """

        try:
            if self.type == "ingress":
                self.compute.cs.revokeSecurityGroupIngress(id=self.id)
            else:
                self.compute.cs.revokeSecurityGroupEgress(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.type = None
        self.id = None
        self.description = None
        self.network_cidr = None
        self.security_group = None
        self.port = None
        self.protocol = None
        self.icmp_code = None
        self.icmp_type = None

    def _parse_port(self):
        """
        Parse the rule port attribute and returns a port range.

        Returns:
            int: start port
            int: end port
        """

        if not self.port:
            return None, None

        start_port, end_port = (
            self.port.split("-", maxsplit=1) if "-" in self.port else (self.port, None)
        )
        if not end_port:
            end_port = start_port

        return int(start_port), int(end_port)


@attr.s
class SSHKey(Resource):
    """
    A SSH key.

    Attributes:
        name (str): the SSH key unique name
        fingerprint (str): the SSH key fingerprint
        private_key (str): the SSH private key, or None if registered SSH key
    """

    compute = attr.ib(repr=False)
    res = attr.ib(repr=False)
    name = attr.ib()
    fingerprint = attr.ib()
    private_key = attr.ib(default=None, repr=False)

    @classmethod
    def _from_cs(cls, compute, res):
        return cls(
            compute,
            res,
            name=res["name"],
            fingerprint=res["fingerprint"],
            private_key=res.get("privatekey", None),
        )

    def delete(self):
        """
        Delete the SSH key.

        Returns:
            None
        """

        try:
            self.compute.cs.deleteSSHKeyPair(name=self.name)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.res = None
        self.name = None
        self.fingerprint = None
        self.private_key = None


@attr.s
class Zone(Resource):
    """
    An Exoscale zone.

    Attributes:
        id (str): the zone unique identifier
        name (str): the zone name
    """

    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()

    @classmethod
    def _from_cs(cls, res):
        return cls(res, id=res["id"], name=res["name"])


class ComputeAPI(API):
    """
    An Exoscale Compute API client.

    Parameters:
        key (str): the Compute API key
        secret (str): the Compute API secret
        endpoint (str): the Compute API endpoint
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API request/response tracing flag
    """

    def __init__(
        self,
        key,
        secret,
        endpoint="https://api.exoscale.com/v1",
        environment="api",
        max_retries=None,
        trace=False,
    ):
        super().__init__(
            endpoint=endpoint,
            key=key,
            secret=secret,
            max_retries=max_retries,
            trace=trace,
        )

        self.environment = environment

        self.cs = CloudStack(
            key=key,
            secret=secret,
            endpoint=endpoint,
            session=self.session,
            headers={**self.session.headers, **{"User-Agent": self.user_agent}},
            trace=self.trace,
            fetch_result=True,
        )

    def __repr__(self):
        return "ComputeAPI(endpoint='{}' key='{}')".format(self.endpoint, self.key)

    def __str__(self):
        return self.__repr__()

    ### Anti-Affinity Group

    def create_anti_affinity_group(self, name, description=""):
        """
        Create an Anti-Affinity Group.

        Parameters:
            name (str): the Anti-Affinity Group name
            description (str): the Anti-Affinity Group description

        Returns:
            AntiAffinityGroup: the Anti-Affinity Group created
        """

        try:
            res = self.cs.createAffinityGroup(
                name=name, description=description, type="host anti-affinity"
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return AntiAffinityGroup._from_cs(self, res["affinitygroup"])

    def list_anti_affinity_groups(self, **kwargs):
        """
        List Anti-Affinity Groups.

        Yields:
            AntiAffinityGroup: the next Anti-Affinity Group
        """

        try:
            _list = self.cs.listAffinityGroups(fetch_list=True, **kwargs)

            for i in _list:
                if i["type"] == "host anti-affinity":
                    yield AntiAffinityGroup._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_anti_affinity_group(self, name=None, id=None):
        """
        Get an Anti-Affinity Group.

        Parameters:
            id (str): an Anti-Affinity Group identifier
            name (str): an Anti-Affinity Group name

        Returns:
            AntiAffinityGroup: an Anti-Affinity Group
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        try:
            anti_affinity_groups = list(
                self.list_anti_affinity_groups(id=id, name=name)
            )
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(anti_affinity_groups) == 0:
            raise ResourceNotFoundError

        return anti_affinity_groups[0]

    ### Elastic IP

    def create_elastic_ip(
        self,
        zone,
        description=None,
        healthcheck_mode=None,
        healthcheck_port=None,
        healthcheck_path="/",
        healthcheck_interval=10,
        healthcheck_timeout=2,
        healthcheck_strikes_ok=3,
        healthcheck_strikes_fail=2,
        healthcheck_tls_sni=None,
        healthcheck_tls_skip_verify=None,
    ):
        """
        Create an Elastic IP.

        Parameters:
            zone (Zone): the zone in which to create the Elastic IP
            description (str): an Elastic IP description
            healthcheck_mode (str): optional healthcheck mode
            healthcheck_port (int): healthcheck port,
                required if healthchecking is enabled
            healthcheck_path (str): healthcheck probe HTTP request path,
                required in "http" mode
            healthcheck_interval (int): probe interval in seconds
            healthcheck_timeout (int): time in seconds before
                considering a probe failed, must be lower than interval
            healthcheck_strikes_ok (int): number of successful probes
                before considering the target healthy
            healthcheck_strikes_fail (int): number of unsuccessful probes
                before considering the target unhealthy
            healthcheck_tls_sni (str): the TLS SNI domain to present for HTTPS
                healthchecks
            healthcheck_tls_skip_verify (bool): whether to skip TLS certificate
                validation for HTTPS healthchecks

        Returns:
            ElasticIP: the Elastic IP created
        """

        # Unset healthcheck default values if healthchecking is not enabled
        if healthcheck_mode is None:
            healthcheck_path = None
            healthcheck_interval = None
            healthcheck_timeout = None
            healthcheck_strikes_ok = None
            healthcheck_strikes_fail = None
            healthcheck_tls_sni = None
            healthcheck_tls_skip_verify = None

        try:
            # We have to pass function arguments using **kwargs form because
            # of the hyphen in arguments names.
            res = self.cs.associateIpAddress(
                **{
                    "zoneid": zone.id,
                    "description": description,
                    "mode": healthcheck_mode,
                    "port": healthcheck_port,
                    "path": healthcheck_path,
                    "interval": healthcheck_interval,
                    "timeout": healthcheck_timeout,
                    "strikes-ok": healthcheck_strikes_ok,
                    "strikes-fail": healthcheck_strikes_fail,
                    "tls-sni": healthcheck_tls_sni,
                    "tls-skip-verify": healthcheck_tls_skip_verify,
                }
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return ElasticIP._from_cs(self, res["ipaddress"], zone=zone)

    def list_elastic_ips(self, zone, **kwargs):
        """
        List Elastic IPs.

        Parameters:
            zone (Zone): the zone to list in

        Yields:
            ElasticIP: the next Elastic IP
        """

        try:
            _list = self.cs.listPublicIpAddresses(
                fetch_list=True, zoneid=zone.id, iselastic=True, **kwargs
            )

            for i in _list:
                yield ElasticIP._from_cs(self, i, zone=zone)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_elastic_ip(self, zone, address=None, id=None):
        """
        Retrieve an Elastic IP.

        Parameters:
            zone (Zone): the zone to retrieve from
            address (str): an Elastic IP address
            id (str): an Elastic IP identifier

        Returns:
            ElasticIP: an Elastic IP
        """

        if id is None and address is None:
            raise ValueError("either id or address must be specifed")

        try:
            elastic_ips = list(self.list_elastic_ips(zone, id=id, ipaddress=address))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(elastic_ips) == 0:
            raise ResourceNotFoundError

        return elastic_ips[0]

    ### Instance

    def create_instance(
        self,
        name,
        zone,
        type,
        template,
        volume_size=10,
        security_groups=None,
        anti_affinity_groups=None,
        private_networks=None,
        enable_ipv6=False,
        ssh_key=None,
        user_data=None,
    ):
        """
        Create a Compute instance.

        Parameters:
            name (str): the name of the instance
            zone (Zone): the zone in which to create the Compute instance
            type (InstanceType): the instance type
            template (InstanceTemplate): the instance template
            volume_size (int): the instance storage volume size in GB
            security_groups ([SecurityGroup]): a list of Security Groups to attach
                the instance to
            anti_affinity_groups ([AntiAffinityGroup]): a list of Anti-Affinity Groups
                to place the instance into
            private_networks ([PrivateNetwork]): a list of Private Networks to attach
                the instance to
            enable_ipv6 (bool): a flag indicating whether to enable IPv6 on the public
                network interface
            ssh_key (SSHKey): a SSH Key to deploy on the instance
            user_data (str): a cloud-init user data configuration

        Returns:
            Instance: the Compute instance created
        """

        try:
            res = self.cs.deployVirtualMachine(
                name=name,
                displayname=name,
                zoneid=zone.id,
                serviceofferingid=type.id,
                templateid=template.id,
                rootdisksize=volume_size,
                securitygroupids=list(i.id for i in security_groups)
                if security_groups
                else None,
                affinitygroupids=list(i.id for i in anti_affinity_groups)
                if anti_affinity_groups
                else None,
                networkids=list(i.id for i in private_networks)
                if private_networks
                else None,
                ip6=enable_ipv6,
                keypair=ssh_key.name if ssh_key else None,
                userdata=b64encode(bytes(user_data, encoding="utf-8"))
                if user_data
                else None,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return Instance._from_cs(self, res["virtualmachine"], zone=zone)

    def list_instances(self, zone, name=None, ids=None, **kwargs):
        """
        List Compute instances.

        Parameters:
            zone (Zone): the zone to list in
            name (str): an Instance name to restrict results to
            ids ([str]): a list of Instance IDs to restrict results to

        Yields:
            Instance: the next Compute instance
        """

        try:
            _list = self.cs.listVirtualMachines(
                fetch_list=True, zoneid=zone.id, ids=ids, name=name, **kwargs
            )

            for i in _list:
                yield Instance._from_cs(self, i, zone=zone)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_instance(self, zone, id=None, ip_address=None):
        """
        Get a Compute instance.

        Parameters:
            zone (Zone): the zone to retrieve from
            id (str): an instance identifier
            ip_address (str): an instance IP address

        Returns:
            Instance: a Compute instance

        Note:
            The ``ip_address`` parameter is a Compute instance's *primary* IP address,
            not an Elastic IP.
        """

        if id is None and ip_address is None:
            raise ValueError("either id or ip_address must be specifed")

        try:
            instances = list(self.list_instances(zone, id=id, ipaddress=ip_address))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(instances) == 0:
            raise ResourceNotFoundError

        return instances[0]

    ### Instance Template

    def register_instance_template(
        self,
        name,
        url,
        checksum,
        zone,
        description=None,
        username=None,
        disable_ssh_key=False,
        disable_password_reset=False,
    ):
        """
        Register a custom instance template.

        Attributes:
            name (str): the instance template name
            description (str): an instance template description
            url (str): the URL at which to find the instance template disk image
            checksum (str): the instance template disk image MD5 checksum
            username (str): an username to log into Compute instances using this
                template
            disable_ssh_key (bool): a flag indicating whether to disable SSH key
                installation during Compute instance creation
            disable_password_reset (bool): a flag indicating whether to disable
                Compute instance password reset
        """
        return InstanceTemplate._register(
            compute=self,
            name=name,
            description=description,
            url=url,
            checksum=checksum,
            zone=zone,
            username=username,
            disable_ssh_key=disable_ssh_key,
            disable_password_reset=disable_password_reset,
        )

    def list_instance_templates(self, zone, name=None, type="exoscale", **kwargs):
        """
        List instance templates.

        Parameters:
            zone (Zone): the zone to list in
            name (str): an instance template name to restrict results to
            type (str): an instance template type to restrict results to

        Yields:
            InstanceTemplate: the next instance template
        """

        template_filters = {"exoscale": "featured", "mine": "self"}
        if type not in template_filters:
            raise ValueError(
                'invalid type "{}", supported types are: {}'.format(
                    type, ", ".join(template_filters)
                )
            )

        try:
            _list = self.cs.listTemplates(
                fetch_list=True,
                zoneid=zone.id,
                name=name,
                templatefilter=template_filters[type],
                **kwargs,
            )

            for i in _list:
                yield InstanceTemplate._from_cs(self, i, zone=zone)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_instance_template(self, zone, id):
        """
        Get an instance template.

        Parameters:
            zone (Zone): the zone to retrieve from
            id (str): an instance template identifier

        Returns:
            InstanceTemplate: an instance template
        """

        try:
            instance_templates = list(self.list_instance_templates(zone, id=id))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(instance_templates) == 0:
            raise ResourceNotFoundError

        return instance_templates[0]

    ### Instance Type

    def list_instance_types(self, **kwargs):
        """
        List Compute instance types.

        Yields:
            InstanceType: the next instance type
        """

        try:
            _list = self.cs.listServiceOfferings(fetch_list=True, **kwargs)

            for i in _list:
                yield InstanceType._from_cs(i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_instance_type(self, name=None, id=None):
        """
        Get a Compute instance type.

        Parameters:
            name (str): an instance type name
            id (str): an instance type identifier

        Returns:
            InstanceType: a Compute instance type
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        try:
            instance_types = list(self.list_instance_types(id=id, name=name))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(instance_types) == 0:
            raise ResourceNotFoundError

        return instance_types[0]

    ### Instance Pool

    def create_instance_pool(
        self,
        zone,
        name,
        size,
        instance_type,
        instance_template,
        instance_volume_size=10,
        instance_anti_affinity_groups=None,
        instance_security_groups=None,
        instance_private_networks=None,
        instance_ssh_key=None,
        instance_user_data=None,
        description=None,
    ):
        """
        Create a Compute Instance Pool.

        Parameters:
            zone (Zone): the zone in which to create the Instance Pool
            name (str): the name of the Instance Pool
            description (str): a description of the Instance Pool
            size (int): the number of Compute instance members the pool must manage
            instance_type (InstanceType): the Compute instance members type
            instance_template (InstanceTemplate): the Compute instance template to use
                for members
            instance_volume_size (int): the Compute instance members storage volume
                size in GB
            instance_private_networks ([PrivateNetwork]): a list of Private Networks to
                attach the Compute instance members to
            instance_anti_affinity_groups ([AntiAffinityGroup]): a list of Anti-Affinity
                Groups to attach the Compute instance members to
            instance_security_groups ([SecurityGroup]): a list of Security Groups to
                attach the Compute instance members to
            instance_ssh_key (SSHKey): a SSH Key to deploy on the Compute instance
                members
            instance_user_data (str): a cloud-init user data configuration to apply to
                the Compute instance members

        Returns:
            InstancePool: the Compute Instance Pool created
        """

        if size <= 0:
            raise ValueError("size must be > 0")

        instance_anti_affinity_group_ids = None
        if instance_anti_affinity_groups:
            instance_anti_affinity_group_ids = [
                i.id for i in instance_anti_affinity_groups
            ]

        instance_security_group_ids = None
        if instance_security_groups:
            instance_security_group_ids = [i.id for i in instance_security_groups]

        instance_private_network_ids = None
        if instance_security_groups:
            instance_private_network_ids = [i.id for i in instance_private_networks]

        instance_ssh_key_name = None
        if instance_ssh_key:
            instance_ssh_key_name = instance_ssh_key.name

        instance_user_data_content = None
        if instance_user_data is not None:
            instance_user_data_content = b64encode(
                bytes(instance_user_data, encoding="utf-8")
            )

        try:
            res = self.cs.createInstancePool(
                name=name,
                description=description,
                size=size,
                zoneid=zone.id,
                serviceofferingid=instance_type.id,
                templateid=instance_template.id,
                rootdisksize=instance_volume_size,
                affinitygroupids=instance_anti_affinity_group_ids,
                securitygroupids=instance_security_group_ids,
                networkids=instance_private_network_ids,
                keypair=instance_ssh_key_name,
                userdata=instance_user_data_content,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return InstancePool._from_cs(self, res, zone=zone)

    def list_instance_pools(self, zone, **kwargs):
        """
        List Compute Instance Pools.

        Parameters:
            zone (Zone): a zone to restrict results to

        Yields:
            InstancePool: the next Compute Instance Pool
        """

        try:
            _list = self.cs.listInstancePools(zoneid=zone.id, fetch_list=True, **kwargs)

            for i in _list:
                yield InstancePool._from_cs(self, i, zone=zone)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_instance_pool(self, zone, id):
        """
        Get an Instance Pool.

        Parameters:
            zone (Zone): the zone in which the Instance Pool is located in
            id (str): an Instance Pool identifier

        Returns:
            InstancePool: an Instance Pool
        """

        try:
            [res] = self.cs.getInstancePool(id=id, zoneid=zone.id, fetch_list=True)
        except CloudStackApiException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            else:
                raise APIException(e.error["errortext"], e.error)

        return InstancePool._from_cs(self, res, zone=zone)

    ### Network Load Balancer

    def create_network_load_balancer(self, zone, name, description=""):
        """
        Create a Network Load Balancer.

        Parameters:
            zone (Zone): the zone in which to create the Network Load Balancer
            name (str): the Network Load Balancer name
            description (str): the Network Load Balancer description

        Returns:
            NetworkLoadBalancer: the Network Load Balancer created
        """

        res = self._v2_request_async(
            "POST",
            "/load-balancer",
            zone=zone.name,
            json={"name": name, "description": description},
        )

        return self.get_network_load_balancer(zone, id=res["reference"]["id"])

    def list_network_load_balancers(self, zone):
        """
        List Network Load Balancers.

        Parameters:
            zone (Zone): the zone to list in

        Yields:
            NetworkLoadBalancer: the next Network Load Balancer
        """

        _list = self._v2_request("GET", "/load-balancer", zone.name)

        for i in _list["load-balancers"]:
            yield NetworkLoadBalancer._from_api(self, i, zone)

    def get_network_load_balancer(self, zone, name=None, id=None):
        """
        Get a Network Load Balancer.

        Parameters:
            name (str): a Network Load Balancer name
            id (str): a Network Load Balancer unique identifier

        Returns:
            NetworkLoadBalancer: a Network Load Balancer
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        for nlb in self.list_network_load_balancers(zone):
            if nlb.id == id or nlb.name == name:
                return nlb

        raise ResourceNotFoundError

    ### Private Network

    def create_private_network(
        self, zone, name, description="", start_ip=None, end_ip=None, netmask=None
    ):
        """
        Create a Private Network.

        Parameters:
            zone (Zone): the zone in which to create the Private Network
            name (str): the Private Network name
            description (str): the Private Network description
            start_ip (str): the network IP range start address for managed
                Private Networks
            end_ip (str): the network IP range end address for managed
                Private Networks
            netmask (str): the network IP netmask for managed Private Networks

        Returns:
            PrivateNetwork: the Private Network created
        """

        try:
            res = self.cs.createNetwork(
                zoneid=zone.id,
                name=name,
                displaytext=description,
                startip=start_ip,
                endip=end_ip,
                netmask=netmask,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return PrivateNetwork._from_cs(self, res["network"], zone=zone)

    def list_private_networks(self, zone, **kwargs):
        """
        List Private Networks.

        Parameters:
            zone (Zone): the zone to list in

        Yields:
            PrivateNetwork: the next Private Network
        """

        try:
            _list = self.cs.listNetworks(fetch_list=True, zoneid=zone.id, **kwargs)

            for i in _list:
                yield PrivateNetwork._from_cs(self, i, zone=zone)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_private_network(self, zone, id):
        """
        Get a Private Network.

        Parameters:
            zone (Zone): the zone to retrieve from
            id (str): a Private Network identifier

        Returns:
            PrivateNetwork: a Private Network
        """

        try:
            private_networks = list(self.list_private_networks(zone, id=id))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(private_networks) == 0:
            raise ResourceNotFoundError

        return private_networks[0]

    ### Security Group

    def create_security_group(self, name, description=""):
        """
        Create a Security Group.

        Parameters:
            name (str): the Security group name
            description (str): the Security Group description

        Returns:
            SecurityGroup: the Security Group created
        """

        try:
            res = self.cs.createSecurityGroup(name=name, description=description)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return SecurityGroup._from_cs(self, res["securitygroup"])

    def list_security_groups(self, **kwargs):
        """
        List Security Groups.

        Yields:
            SecurityGroup: the next Security Group
        """

        try:
            _list = self.cs.listSecurityGroups(fetch_list=True, **kwargs)

            for i in _list:
                yield SecurityGroup._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_security_group(self, name=None, id=None):
        """
        Get a Security Group.

        Parameters:
            name (str): a Security Group name
            id (str): a Security Group identifier

        Returns:
            SecurityGroup: a Security Group
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        try:
            security_groups = list(
                self.list_security_groups(id=id, securitygroupname=name)
            )
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(security_groups) == 0:
            raise ResourceNotFoundError

        return security_groups[0]

    ### SSH Key

    def create_ssh_key(self, name):
        """
        Create an SSH key.

        Parameters:
            name (str): the SSH key unique name

        Returns:
            SSHKey: the SSH key created
        """

        try:
            res = self.cs.createSSHKeyPair(name=name)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return SSHKey._from_cs(self, res["keypair"])

    def register_ssh_key(self, name, public_key):
        """
        Register an existing SSH key.

        Parameters:
            name (str): the SSH Key unique name
            public_key (str): the SSH public key to register

        Returns:
            SSHKey: the SSH key created
        """

        try:
            res = self.cs.registerSSHKeyPair(name=name, publickey=public_key)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return SSHKey._from_cs(self, res["keypair"])

    def list_ssh_keys(self, **kwargs):
        """
        List SSH keys.

        Yields:
            SSHKey: the next SSH key
        """

        try:
            _list = self.cs.listSSHKeyPairs(fetch_list=True, **kwargs)

            for i in _list:
                yield SSHKey._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_ssh_key(self, name):
        """
        Get an SSH Key.

        Parameters:
            name (str): an SSH Key name

        Returns:
            SSHKey: an SSH Key
        """

        try:
            ssh_keys = list(self.list_ssh_keys(name=name))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(ssh_keys) == 0:
            raise ResourceNotFoundError

        return ssh_keys[0]

    ### Zone

    def list_zones(self, **kwargs):
        """
        List zones.

        Yields:
            Zone: the next zone
        """

        try:
            _list = self.cs.listZones(fetch_list=True, **kwargs)

            for i in _list:
                yield Zone._from_cs(i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_zone(self, name=None, id=None):
        """
        Get a zone.

        Parameters:
            name (str): a zone name
            id (str): a zone identifier

        Returns:
            Zone: a zone
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        try:
            zones = list(self.list_zones(id=id, name=name))
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        if len(zones) == 0:
            raise ResourceNotFoundError

        return zones[0]

    ### V2 API

    def _v2_check_response(self, res, *args, **kwargs):
        """
        Check the API response and raise an exception depending on the status code.
        """

        if res.status_code >= 500:
            raise APIException(res.text)

        if res.status_code == 404:
            raise ResourceNotFoundError

        if res.status_code >= 400:
            raise RequestError(str(res.text))

    def _v2_request(self, method, path, zone=None, **kwargs):
        base_url = "https://api.exoscale/v2.alpha"
        if zone:
            base_url = "https://{}-{}.exoscale.com/v2.alpha".format(
                self.environment, zone
            )

        return API.send(
            self,
            method=method,
            url="/".join((base_url, path.lstrip("/"))),
            auth=ExoscaleV2Auth(self.key, self.secret),
            hooks={"response": self._v2_check_response},
            **kwargs,
        ).json()

    def _v2_request_async(self, method, path, zone, **kwargs):
        op = self._v2_request(method, path, zone, **kwargs)

        return polling.poll(
            lambda: self._v2_request("GET", "/operation/" + op["id"], zone),
            check_success=self._v2_check_async_operation_state,
            step=3,
            poll_forever=True,
        )

    def _v2_check_async_operation_state(self, op):
        if op["state"] == "pending":
            return False
        if op["state"] == "success":
            return True
        elif op["state"] == "failure":
            raise APIException("asynchronous operation failed")
        elif op["state"] == "timeout":
            raise APIException("asynchronous operation timed out")
        else:
            raise APIException('unknown operation state "{}"'.format(op.state))



# -*- coding: utf-8 -*-

"""
This submodule represents the Exoscale DNS API.
"""

import attr
from . import API, Resource, APIException, ResourceNotFoundError
from cs import CloudStack, CloudStackApiException

_SUPPORTED_RECORD_TYPES = {
    "A",
    "AAAA",
    "ALIAS",
    "CAA",
    "CNAME",
    "DNSKEY",
    "DS",
    "HINFO",
    "MX",
    "NAPTR",
    "NS",
    "POOL",
    "PTR",
    "SOA",
    "SPF",
    "SRV",
    "SSHFP",
    "TXT",
    "URL",
}


@attr.s
class Domain(Resource):
    """
    A DNS domain.

    Attributes:
        id (int): the DNS domain unique identifier
        name (str): the DNS domain name
    """

    dns = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    unicode_name = attr.ib(repr=False)

    @classmethod
    def _from_cs(cls, dns, res):
        return cls(
            dns, res, id=res["id"], name=res["name"], unicode_name=res["unicodename"]
        )

    @property
    def records(self):
        """
        Domain records.

        Yields:
            DomainRecord: the next DNS domain record defined

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            _list = self.dns.cs.listDnsDomainRecords(id=self.id, fetch_list=True)
            for i in _list:
                yield DomainRecord._from_cs(self.dns, domain=self, res=i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def add_record(self, name, type, content, priority=None, ttl=3600):
        """
        Add a record to the DNS domain.

        Parameters:
            name (str): the DNS domain record name
            type (str): the DNS domain record type
            content (str): the DNS domain record content
            priority (int): the DNS domain record priority
            ttl (int): the DNS domain record TTL

        Returns:
            None
        """

        if type not in _SUPPORTED_RECORD_TYPES:
            raise ValueError(
                "unsupported record type; supported types are: {}".format(
                    ",".join(_SUPPORTED_RECORD_TYPES)
                )
            )

        try:
            self.dns.cs.createDnsDomainRecord(
                name=self.name,
                record_name=name,
                record_type=type,
                content=content,
                priority=priority,
                ttl=ttl,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def delete(self):
        """
        Delete the DNS domain.

        Returns:
            None
        """

        try:
            self.dns.cs.deleteDnsDomain(id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.dns = None
        self.res = None
        self.id = None
        self.name = None
        self.unicode_name = None


@attr.s
class DomainRecord(Resource):
    """
    A DNS domain record.

    Attributes:
        id (int): the DNS domain record unique identifier
        domain (Domain): the DNS domain the record belongs to
        type (str): the DNS domain record type
        name (str): the DNS domain record name
        content (str): the DNS domain record content
        priority (int): the DNS domain record priority
        ttl (int): the DNS domain record TTL
    """

    dns = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    domain = attr.ib(repr=False)
    type = attr.ib()
    name = attr.ib(repr=False)
    content = attr.ib(repr=False)
    priority = attr.ib(default=None, repr=False)
    ttl = attr.ib(default=3600, repr=False)

    @classmethod
    def _from_cs(cls, dns, res, domain):
        return cls(
            dns,
            res,
            domain=domain,
            id=res["id"],
            type=res["record_type"],
            name=res["name"],
            content=res["content"],
            priority=res["priority"],
            ttl=res["ttl"],
        )

    def update(self, name=None, content=None, priority=None, ttl=3600):
        """
        Update a DNS domain record.

        Parameters:
            name (str): the DNS domain record name
            content (str): the DNS domain record content
            priority (int): the DNS domain record priority
            ttl (int): the DNS domain record TTL

        Returns:
            None
        """

        try:
            self.dns.cs.updateDnsDomainRecord(
                name=self.domain.name,
                record_id=self.id,
                record_type=self.type,
                record_name=name if name else self.name,
                content=content if content else self.content,
                priority=priority if priority else self.priority,
                ttl=ttl if ttl else self.ttl,
            )
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.name = name if name else self.name
        self.content = content if content else self.content
        self.priority = priority if priority else self.priority
        self.ttl = ttl if ttl else self.ttl

    def delete(self):
        """
        Delete the DNS domain record.

        Returns:
            None
        """

        try:
            self.dns.cs.deleteDnsDomainRecord(id=self.domain.id, record_id=self.id)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.dns = None
        self.res = None
        self.id = None
        self.domain = None
        self.type = None
        self.name = None
        self.content = None
        self.priority = None
        self.ttl = None


class DnsAPI(API):
    """
    An Exoscale DNS API client.

    Parameters:
        key (str): the DNS API key
        secret (str): the DNS API secret
        endpoint (str): the DNS API endpoint
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API request/response tracing flag
    """

    def __init__(
        self,
        key,
        secret,
        endpoint="https://api.exoscale.com/v1",
        max_retries=None,
        trace=False,
    ):
        super().__init__(
            endpoint=endpoint,
            key=key,
            secret=secret,
            max_retries=max_retries,
            trace=trace,
        )

        self.cs = CloudStack(
            key=key,
            secret=secret,
            endpoint=endpoint,
            session=self.session,
            headers={**self.session.headers, **{"User-Agent": self.user_agent}},
            trace=self.trace,
            fetch_result=True,
        )

    def __repr__(self):
        return "DnsAPI(endpoint='{}' key='{}')".format(self.endpoint, self.key)

    def __str__(self):
        return self.__repr__()

    ### Domain

    def create_domain(self, name):
        """
        Create a DNS domain.

        Parameters:
            name (str): the domain name

        Returns:
            Domain: the DNS domain created
        """

        try:
            res = self.cs.createDnsDomain(name=name)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return Domain._from_cs(self, res["dnsdomain"])

    def list_domains(self):
        """
        List DNS domains.

        Yields:
            Domain: the next DNS domain
        """

        try:
            _list = self.cs.listDnsDomains(fetch_list=True)

            for i in _list:
                yield Domain._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_domain(self, name=None, id=None):
        """
        Get a DNS domain.

        Parameters:
            id (str): a DNS domain identifier
            name (str): a DNS domain name

        Returns:
            Domain: a DNS domain
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        try:
            domains = self.list_domains()
        except APIException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            raise

        for domain in domains:
            if (name and domain.name == name) or (id and domain.id == id):
                return domain

        raise ResourceNotFoundError



# -*- coding: utf-8 -*-

"""
This submodule represents the Exoscale IAM API.
"""

import attr
from . import API, Resource, APIException, ResourceNotFoundError
from cs import CloudStack, CloudStackApiException


@attr.s
class APIKey(Resource):
    """
    An API key.

    Attributes:
        name (str): the API key display name
        key (str): the API key unique identifier
        type (str): the API key type
        secret (str): the API key secret
        operations ([str]): a list of allowed API operations
    """

    iam = attr.ib(repr=False)
    res = attr.ib(repr=False)
    name = attr.ib()
    key = attr.ib()
    type = attr.ib()
    secret = attr.ib(repr=False)
    operations = attr.ib(repr=False)

    @classmethod
    def _from_cs(cls, iam, res):
        return cls(
            iam,
            res,
            name=res["name"],
            key=res["key"],
            type=res["type"],
            secret=res.get("secret", None),
            operations=res.get("operations", None),
        )

    def revoke(self):
        """
        Revoke the API key.

        Returns:
            None
        """

        try:
            self.iam.cs.revokeApiKey(key=self.key)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        self.res = None
        self.name = None
        self.key = None
        self.secret = None
        self.operations = None
        self.type = None


class IamAPI(API):
    """
    An Exoscale IAM API client.

    Parameters:
        key (str): the API key unique identifier
        secret (str): the IAM API secret
        endpoint (str): the IAM API endpoint
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API request/response tracing flag
    """

    def __init__(
        self,
        key,
        secret,
        endpoint="https://api.exoscale.com/compute",
        max_retries=None,
        trace=False,
    ):
        super().__init__(
            endpoint=endpoint,
            key=key,
            secret=secret,
            max_retries=max_retries,
            trace=trace,
        )

        self.cs = CloudStack(
            key=key,
            secret=secret,
            endpoint=endpoint,
            session=self.session,
            headers={**self.session.headers, **{"User-Agent": self.user_agent}},
            trace=self.trace,
            fetch_result=True,
        )

    def __repr__(self):
        return "IamAPI(endpoint='{}' key='{}')".format(self.endpoint, self.key)

    def __str__(self):
        return self.__repr__()

    ### API key

    def create_api_key(self, name, operations=None):
        """
        Create an API key.

        Parameters:
            name (str): the API key name
            operations (str): a comma-separated list of allowed API operations

        Returns:
           APIKey: the API key created
        """

        try:
            res = self.cs.createApiKey(name=name, operations=operations)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

        return APIKey._from_cs(self, res["apikey"])

    def list_api_keys(self, **kwargs):
        """
        List API keys.

        Yields:
           APIKey: the next API key
        """

        try:
            _list = self.cs.listApiKeys(fetch_list=True, **kwargs)

            for i in _list:
                yield APIKey._from_cs(self, i)
        except CloudStackApiException as e:
            raise APIException(e.error["errortext"], e.error)

    def get_api_key(self, key):
        """
        Get an API key.

        Parameters:
            key (str): the API key unique identifier

        Returns:
            APIKey: an API key
        """

        try:
            res = self.cs.getApiKey(key=key)
        except CloudStackApiException as e:
            if "The key is not found." in e.error["errortext"]:
                raise ResourceNotFoundError
            else:
                raise CloudStackApiException(e.error["errortext"], e.error)

        return APIKey._from_cs(self, res)

    def list_api_key_operations(self):
        """
         list all supported operations of an API key.

        Returns:
             [str]: list of operations for the current API key
        """

        try:
            res = self.cs.listApiKeyOperations()
        except CloudStackApiException as e:
            if "does not exist" in e.error["errortext"]:
                raise ResourceNotFoundError
            else:
                raise APIException(e.error["errortext"], e.error)

        return res["operations"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.runstatus import *


class TestRunstatusPageService:
    def test_delete(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        service_name = "a"

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name), json={"name": service_name}
        )

        res = exo.runstatus._get(url="/pages/{p}/services".format(p=page.name)).json()
        service = Service._from_rs(exo.runstatus, res["results"][0], page=page)

        service.delete()
        assert service.id == None

        res = exo.runstatus._get(url="/pages/{p}/services".format(p=page.name)).json()
        assert len(res["results"]) == 0

    def test_properties(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        service_name = "a"

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name), json={"name": service_name}
        )

        res = exo.runstatus._get(url="/pages/{p}/services".format(p=page.name)).json()
        service = Service._from_rs(exo.runstatus, res["results"][0], page=page)

        assert service.state == "operational"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from cs import CloudStackApiException
from exoscale.api.dns import *


class TestDNSDomain:
    def test_add_record(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain())
        domain_record_name = "test-python-exoscale"
        domain_record_type = "MX"
        domain_record_content = "mx1.example.net"
        domain_record_priority = 10
        domain_record_ttl = 1042

        domain.add_record(
            name=domain_record_name,
            type=domain_record_type,
            content=domain_record_content,
            priority=domain_record_priority,
            ttl=domain_record_ttl,
        )

        res = exo.dns.cs.listDnsDomainRecords(id=domain.id, fetch_list=True)
        for record in res:
            if record["name"] == "":
                continue
            assert record["name"] == domain_record_name
            assert record["record_type"] == domain_record_type
            assert record["content"] == domain_record_content
            assert record["priority"] == domain_record_priority
            assert record["ttl"] == domain_record_ttl

    def test_delete(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain(teardown=False))
        domain_id = domain.id

        domain.delete()
        assert domain.id == None

        res = exo.dns.cs.listDnsDomains(fetch_list=True)
        for i in res:
            assert i["id"] != domain_id

    def test_properties(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain())
        domain_record_name = "test-python-exoscale"
        domain_record_type = "A"
        domain_record_content = "1.2.3.4"
        domain_record_ttl = 1042

        exo.dns.cs.createDnsDomainRecord(
            name=domain.name,
            record_name=domain_record_name,
            record_type=domain_record_type,
            content=domain_record_content,
            ttl=domain_record_ttl,
        )

        domain_records = list(domain.records)
        assert len(domain_records) > 1
        domain_record = list(r for r in domain_records if r.name == domain_record_name)[
            0
        ]
        assert domain_record.id > 0
        assert domain_record.type == domain_record_type
        assert domain_record.content == domain_record_content
        assert domain_record.ttl == domain_record_ttl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.storage import *
from .conftest import _random_str


class TestStorage:
    ### Bucket

    def test_create_bucket(self, exo, test_prefix):
        bucket_name = "-".join([test_prefix, _random_str()])

        bucket = exo.storage.create_bucket(name=bucket_name)
        assert bucket.name == bucket_name

        exo.storage.boto.delete_bucket(Bucket=bucket_name)

    def test_list_buckets(self, exo, bucket):
        bucket = bucket()

        buckets = list(exo.storage.list_buckets())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture bucket
        assert len(buckets) >= 1

    def test_get_bucket(self, exo, bucket):
        bucket1 = Bucket(exo.storage, {}, bucket())

        bucket = exo.storage.get_bucket(name=bucket1.name)
        assert bucket.name == bucket1.name

        with pytest.raises(ResourceNotFoundError) as excinfo:
            bucket = exo.storage.get_bucket(name="lolnope")
            assert bucket is None
        assert excinfo.type == ResourceNotFoundError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pytest
import random
import string
import time
from datetime import datetime

_DEFAULT_ZONE_NAME = "ch-gva-2"
_DEFAULT_ZONE_ID = "1128bd56-b4d9-4ac6-a7b9-c715b187ce11"


def _random_str():
    chars = string.ascii_lowercase + string.digits
    return "".join(random.choice(chars) for i in range(10))


@pytest.fixture(autouse=True, scope="module")
def test_prefix():
    return "test-python-exoscale"


@pytest.fixture(autouse=True, scope="module")
def test_description():
    return "Created by the python-exoscale library"


@pytest.fixture(autouse=True, scope="module")
def test_instance_service_offering_id():
    # Micro
    return "71004023-bb72-4a97-b1e9-bc66dfce9470"


@pytest.fixture(autouse=True, scope="module")
def test_instance_template_name():
    return "Linux Ubuntu 18.04 LTS 64-bit"


@pytest.fixture(autouse=True, scope="module")
def test_instance_template_id():
    # Linux Ubuntu 18.04 LTS 64-bit @ ch-gva-2
    return "45346aba-6027-45bc-ad1e-bd1f563c2d84"


@pytest.fixture(autouse=True, scope="module")
def test_instance_user_data():
    return "#cloud-config"


@pytest.fixture(autouse=True, scope="module")
def test_reverse_dns():
    return "python.exoscale.com."


@pytest.fixture(autouse=True, scope="function")
def timing():
    print(
        "\n>>> {}: test started at {}".format(
            os.environ.get("PYTEST_CURRENT_TEST"), datetime.now()
        )
    )
    started = time.time()

    yield

    print(
        "\n<<< {}: test stopped at {}".format(
            os.environ.get("PYTEST_CURRENT_TEST"), datetime.now()
        )
    )

    print("--- Test duration: {:0.2f}s\n".format(time.time() - started))


@pytest.fixture(autouse=True, scope="function")
def exo():
    import exoscale

    return exoscale.Exoscale()


@pytest.fixture(autouse=True, scope="function")
def zone(exo):
    def _zone(name=_DEFAULT_ZONE_NAME):
        return exo.compute.cs.listZones(name=name, fetch_list=True)[0]

    return _zone


@pytest.fixture(autouse=True, scope="function")
def instance_type(exo):
    def _instance_type(id=None, name=None):
        return exo.compute.cs.listServiceOfferings(id=id, name=name, fetch_list=True)[0]

    return _instance_type


@pytest.fixture(autouse=True, scope="function")
def instance_template(exo):
    def _instance_template(id=None, name=None, zone_id=None):
        return exo.compute.cs.listTemplates(
            id=id, name=name, zoneid=zone_id, fetch_list=True
        )[0]

    return _instance_template


@pytest.fixture(autouse=True, scope="function")
def aag(exo, test_prefix, test_description):
    anti_affinity_groups = []

    def _anti_affinity_group(name=None, description=test_description, teardown=True):
        anti_affinity_group = exo.compute.cs.createAffinityGroup(
            name=name if name else "-".join([test_prefix, _random_str()]),
            description=description,
            type="host anti-affinity",
        )["affinitygroup"]

        if teardown:
            anti_affinity_groups.append(anti_affinity_group)

        return anti_affinity_group

    yield _anti_affinity_group

    for anti_affinity_group in anti_affinity_groups:
        res = exo.compute.cs.deleteAffinityGroup(id=anti_affinity_group["id"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def eip(exo, test_description):
    elastic_ips = []

    def _elastic_ip(
        zone_id=_DEFAULT_ZONE_ID, description=test_description, teardown=True
    ):
        elastic_ip = exo.compute.cs.associateIpAddress(
            zoneid=zone_id, description=description
        )["ipaddress"]

        if teardown:
            elastic_ips.append(elastic_ip)

        return elastic_ip

    yield _elastic_ip

    for elastic_ip in elastic_ips:
        res = exo.compute.cs.disassociateIpAddress(id=elastic_ip["id"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def instance(
    exo,
    test_prefix,
    test_instance_service_offering_id,
    test_instance_template_id,
):
    instances = []

    def _instance(
        name=None,
        type_id=test_instance_service_offering_id,
        template_id=test_instance_template_id,
        zone_id=_DEFAULT_ZONE_ID,
        volume_size=10,
        security_groups=None,
        anti_affinity_groups=None,
        private_networks=None,
        start=True,
        teardown=True,
    ):
        instance = exo.compute.cs.deployVirtualMachine(
            name=name if name else "-".join([test_prefix, _random_str()]),
            displayname=name,
            zoneid=zone_id,
            templateid=template_id,
            serviceofferingid=type_id,
            root_disk_size=volume_size,
            securitygroupids=security_groups,
            affinitygroupids=anti_affinity_groups,
            networkids=private_networks,
            startvm=start,
        )["virtualmachine"]

        if teardown:
            instances.append(instance)

        return instance

    yield _instance

    for instance in instances:
        exo.compute.cs.destroyVirtualMachine(id=instance["id"])


@pytest.fixture(autouse=True, scope="function")
def instance_pool(
    exo,
    test_prefix,
    test_description,
    test_instance_service_offering_id,
    test_instance_template_id,
    test_instance_user_data,
):
    instance_pools = []

    def _instance_pool(
        zone_id=_DEFAULT_ZONE_ID,
        name=None,
        description=test_description,
        size=1,
        instance_type_id=test_instance_service_offering_id,
        instance_template_id=test_instance_template_id,
        security_groups=None,
        anti_affinity_groups=None,
        private_networks=None,
        teardown=True,
    ):
        instance_pool = exo.compute.cs.createInstancePool(
            zoneid=zone_id,
            name=name if name else "-".join([test_prefix, _random_str()]),
            description=description,
            size=size,
            serviceofferingid=instance_type_id,
            templateid=instance_template_id,
            securitygroupids=security_groups,
            affinitygroupids=anti_affinity_groups,
            networkids=private_networks,
            user_data=test_instance_user_data,
        )

        if teardown:
            instance_pools.append(instance_pool)

        return instance_pool

    yield _instance_pool

    for instance_pool in instance_pools:
        exo.compute.cs.destroyInstancePool(
            id=instance_pool["id"], zoneid=instance_pool["zoneid"]
        )


@pytest.fixture(autouse=True, scope="function")
def nlb(exo, test_prefix, test_description):
    nlbs = []

    def _nlb(
        name=None,
        description=test_description,
        zone_name=_DEFAULT_ZONE_NAME,
        teardown=True,
    ):
        res = exo.compute._v2_request_async(
            "POST",
            "/load-balancer",
            zone=zone_name,
            json={
                "name": name if name else "-".join([test_prefix, _random_str()]),
                "description": description,
            },
        )
        nlb = exo.compute._v2_request(
            "GET", "/load-balancer/" + res["reference"]["id"], zone_name
        )
        nlb["zone"] = zone_name  # Saving NLB zone for later use in teardown

        if teardown:
            nlbs.append(nlb)

        return nlb

    yield _nlb

    for nlb in nlbs:
        res = exo.compute._v2_request_async(
            "DELETE", "/load-balancer/" + nlb["id"], nlb["zone"]
        )


@pytest.fixture(autouse=True, scope="function")
def privnet(exo, test_prefix, test_description):
    private_networks = []

    def _private_network(
        name=None,
        description=test_description,
        zone_id=_DEFAULT_ZONE_ID,
        start_ip=None,
        end_ip=None,
        netmask=None,
        teardown=True,
    ):
        private_network = exo.compute.cs.createNetwork(
            zoneid=zone_id,
            name=name if name else "-".join([test_prefix, _random_str()]),
            displaytext=description,
            startip=start_ip,
            endip=end_ip,
            netmask=netmask,
        )["network"]

        if teardown:
            private_networks.append(private_network)

        return private_network

    yield _private_network

    for private_network in private_networks:
        res = exo.compute.cs.deleteNetwork(id=private_network["id"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def sg(exo, test_prefix, test_description):
    security_groups = []

    def _security_group(name=None, description=test_description, teardown=True):
        security_group = exo.compute.cs.createSecurityGroup(
            name=name if name else "-".join([test_prefix, _random_str()]),
            description=description,
        )["securitygroup"]

        if teardown:
            security_groups.append(security_group)

        return security_group

    yield _security_group

    for security_group in security_groups:
        # We check if there are instances still using the SG before trying to deleteing
        # it: since it's not possible to delete a referenced SG we wait for a while
        # before trying and hope for the best...
        res = exo.compute.cs.deleteSecurityGroup(id=security_group["id"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def sshkey(exo, test_prefix):
    ssh_keys = []

    def _ssh_key(name=None, teardown=True):
        ssh_key = exo.compute.cs.createSSHKeyPair(
            name=name if name else "-".join([test_prefix, _random_str()])
        )["keypair"]

        if teardown:
            ssh_keys.append(ssh_key)

        return ssh_key

    yield _ssh_key

    for ssh_key in ssh_keys:
        res = exo.compute.cs.deleteSSHKeyPair(name=ssh_key["name"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def domain(exo, test_prefix):
    domains = []

    def _domain(name=None, teardown=True):
        domain = exo.dns.cs.createDnsDomain(
            name="-".join([test_prefix, _random_str()]) + ".net"
        )["dnsdomain"]

        if teardown:
            domains.append(domain)

        return domain

    yield _domain

    for domain in domains:
        res = exo.dns.cs.deleteDnsDomain(id=domain["id"])
        assert res["success"]


@pytest.fixture(autouse=True, scope="function")
def bucket(exo, test_prefix):
    buckets = []

    def _bucket(name=None, zone=None, acl="private", teardown=True):
        bucket = exo.storage.boto.create_bucket(
            Bucket=name if name else "-".join([test_prefix, _random_str()]),
            CreateBucketConfiguration={
                "LocationConstraint": _DEFAULT_ZONE_NAME if zone is None else zone
            },
            ACL=acl,
        )["Location"].lstrip("/")

        if teardown:
            buckets.append(bucket)

        return bucket

    yield _bucket

    for bucket in buckets:
        res = exo.storage.boto.delete_bucket(Bucket=bucket)


@pytest.fixture(autouse=True, scope="function")
def runstatus_page(exo, test_prefix):
    pages = []

    def _page(name=None, teardown=True):
        name = name if name else "-".join([test_prefix, _random_str()])
        page = exo.runstatus._post(
            url="/pages", json={"name": name, "subdomain": name}
        ).json()

        if teardown:
            pages.append(page)

        return page

    yield _page

    for page in pages:
        res = exo.runstatus._delete(url="/pages/{p}".format(p=page["subdomain"]))


@pytest.fixture(autouse=True, scope="function")
def apikey(exo, test_prefix):
    api_keys = []

    def _api_key(name=None, teardown=True):
        api_key = exo.iam.cs.createApiKey(
            name=name if name else "-".join([test_prefix, _random_str()])
        )["apikey"]

        if teardown:
            api_keys.append(api_key)

        return api_key

    yield _api_key

    for api_key in api_keys:
        res = exo.iam.cs.revokeApiKey(key=api_key["key"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.dns import *
from .conftest import _random_str


class TestDNS:
    ### Domain

    def test_create_domain(self, exo, test_prefix):
        domain_name = "-".join([test_prefix, _random_str()]) + ".net"

        domain = exo.dns.create_domain(name=domain_name)
        assert domain.id > 0
        assert domain.name == domain_name

        exo.dns.cs.deleteDnsDomain(id=domain.id)

    def test_list_domains(self, exo, domain):
        domain = Domain._from_cs(exo.dns, domain())

        domains = list(exo.dns.list_domains())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture domain
        assert len(domains) >= 1

    def test_get_domain(self, exo, domain):
        domain1 = Domain._from_cs(exo.dns, domain())

        domain = exo.dns.get_domain(id=domain1.id)
        assert domain.id == domain1.id

        domain = exo.dns.get_domain(name=domain1.name)
        assert domain.id == domain1.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            domain = exo.dns.get_domain(id=42)
            assert domain is None
        assert excinfo.type == ResourceNotFoundError

        with pytest.raises(ResourceNotFoundError) as excinfo:
            domain = exo.dns.get_domain(name="lolnope.com")
            assert domain is None
        assert excinfo.type == ResourceNotFoundError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *
from exoscale.api import ResourceNotFoundError
from .conftest import _random_str


class TestComputeNetworkLoadBalancer:
    def test_add_service(
        self, exo, zone, instance_pool, nlb, test_prefix, test_description
    ):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )
        nlb = NetworkLoadBalancer._from_api(exo.compute, nlb(), zone)
        service_name = "-".join([test_prefix, _random_str()])
        service_description = test_description
        service_port = 443
        service_target_port = 8443
        service_protocol = "tcp"
        service_strategy = "source-hash"
        service_healthcheck_mode = "https"
        service_healthcheck_port = 8444
        service_healthcheck_uri = "/health"
        service_healthcheck_interval = 5
        service_healthcheck_timeout = 3
        service_healthcheck_retries = 1
        service_healthcheck_tls_sni = "example.net"

        service = nlb.add_service(
            name=service_name,
            description=service_description,
            instance_pool=instance_pool,
            port=service_port,
            target_port=service_target_port,
            protocol=service_protocol,
            strategy=service_strategy,
            healthcheck_mode=service_healthcheck_mode,
            healthcheck_port=service_healthcheck_port,
            healthcheck_uri=service_healthcheck_uri,
            healthcheck_interval=service_healthcheck_interval,
            healthcheck_timeout=service_healthcheck_timeout,
            healthcheck_retries=service_healthcheck_retries,
            healthcheck_tls_sni=service_healthcheck_tls_sni,
        )

        actual_service = exo.compute._v2_request(
            "GET",
            "/load-balancer/{}/service/{}".format(nlb.id, service.id),
            zone=zone.name,
        )

        assert service.id == actual_service["id"]
        assert service.name == service_name
        assert actual_service["name"] == service_name
        assert service.description == test_description
        assert actual_service["description"] == test_description
        assert service.instance_pool.id == instance_pool.id
        assert actual_service["instance-pool"]["id"] == instance_pool.id
        assert service.port == service_port
        assert actual_service["port"] == service_port
        assert service.target_port == service_target_port
        assert actual_service["target-port"] == service_target_port
        assert service.protocol == service_protocol
        assert actual_service["protocol"] == service_protocol
        assert service.strategy == service_strategy
        assert actual_service["strategy"] == service_strategy
        assert service.healthcheck.mode == service_healthcheck_mode
        assert actual_service["healthcheck"]["mode"] == service_healthcheck_mode
        assert service.healthcheck.port == service_healthcheck_port
        assert actual_service["healthcheck"]["port"] == service_healthcheck_port
        assert service.healthcheck.uri == service_healthcheck_uri
        assert actual_service["healthcheck"]["uri"] == service_healthcheck_uri
        assert service.healthcheck.interval == service_healthcheck_interval
        assert actual_service["healthcheck"]["interval"] == service_healthcheck_interval
        assert service.healthcheck.timeout == service_healthcheck_timeout
        assert actual_service["healthcheck"]["timeout"] == service_healthcheck_timeout
        assert service.healthcheck.retries == service_healthcheck_retries
        assert actual_service["healthcheck"]["retries"] == service_healthcheck_retries
        assert service.healthcheck.tls_sni == service_healthcheck_tls_sni
        assert actual_service["healthcheck"]["tls-sni"] == service_healthcheck_tls_sni

    def test_update(self, exo, zone, nlb):
        zone = Zone._from_cs(zone("ch-gva-2"))
        nlb = NetworkLoadBalancer._from_api(exo.compute, nlb(), zone)
        name_edited = nlb.name + "-edited"
        description_edited = nlb.description + " (edited)"

        nlb.update(
            name=name_edited,
            description=description_edited,
        )

        res = exo.compute._v2_request("GET", "/load-balancer/" + nlb.id, zone=zone.name)
        assert res["name"] == name_edited
        assert nlb.name == name_edited
        assert res["description"] == description_edited
        assert nlb.description == description_edited

    def test_delete(self, exo, zone, nlb):
        zone = Zone._from_cs(zone("ch-gva-2"))
        nlb = NetworkLoadBalancer._from_api(exo.compute, nlb(teardown=False), zone)
        nlb_id = nlb.id

        nlb.delete()
        assert nlb.id is None

        with pytest.raises(ResourceNotFoundError) as excinfo:
            res = exo.compute._v2_request("GET", "/load-balancer/" + nlb_id, zone.name)
            assert res is None
        assert excinfo.type == ResourceNotFoundError

    def test_properties(self, exo, zone, instance_pool, nlb):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )
        nlb = NetworkLoadBalancer._from_api(exo.compute, nlb(), zone)

        res = exo.compute._v2_request_async(
            "POST",
            "/load-balancer/{}/service".format(nlb.id),
            zone.name,
            json={
                "name": "test",
                "instance-pool": {"id": instance_pool.id},
                "port": 80,
                "target-port": 80,
                "protocol": "tcp",
                "strategy": "round-robin",
                "healthcheck": {
                    "mode": "tcp",
                    "port": 80,
                    "interval": 10,
                    "timeout": 5,
                    "retries": 1,
                },
            },
        )

        service = NetworkLoadBalancerService._from_api(
            exo.compute,
            exo.compute._v2_request("GET", "/load-balancer/" + nlb.id, zone.name)[
                "services"
            ][0],
            nlb,
        )
        assert service is not None

        nlb_services = list(nlb.services)
        assert len(nlb_services) == 1
        assert nlb_services[0].id == service.id

        nlb_state = nlb.state
        assert nlb_state == "running"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from datetime import datetime
from exoscale.api.storage import *
from time import sleep


class TestStorageBucketFile:
    def test_delete(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")
        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_a), Key="a")

        f = BucketFile(exo.storage, {}, path="a", bucket=bucket)
        f.delete()

        res = exo.storage.boto.list_objects(Bucket=bucket.name)
        assert "Contents" not in res

    def test_set_acl(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")
        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_a), Key="a")

        f = BucketFile(exo.storage, {}, path="a", bucket=bucket)
        f.set_acl(acl="public-read")
        res = exo.storage.boto.get_object_acl(Bucket=bucket.name, Key="a")
        for i in res["Grants"]:
            if i["Permission"] == "READ":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_ALL_USERS

        f_acp = AccessControlPolicy._from_s3(res)
        f_acp.read = "ALL_USERS"
        f_acp.write = "AUTHENTICATED_USERS"
        f_acp.read_acp = "alice@example.net"
        f_acp.write_acp = "bob@example.net"
        f.set_acl(acp=f_acp)
        res = exo.storage.boto.get_object_acl(Bucket=bucket.name, Key="a")
        for i in res["Grants"]:
            if i["Permission"] == "READ":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_ALL_USERS
            if i["Permission"] == "WRITE":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_AUTHENTICATED_USERS
            if i["Permission"] == "READ_ACP":
                assert i["Grantee"]["Type"] == "CanonicalUser"
                assert i["Grantee"]["ID"] == "alice@example.net"
                assert i["Grantee"]["DisplayName"] == "alice@example.net"
            if i["Permission"] == "WRITE_ACP":
                assert i["Grantee"]["Type"] == "CanonicalUser"
                assert i["Grantee"]["ID"] == "bob@example.net"
                assert i["Grantee"]["DisplayName"] == "bob@example.net"

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="a")

    def test_properties(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")
        exo.storage.boto.upload_file(
            Bucket=bucket.name,
            Filename=str(f_a),
            Key="a",
            ExtraArgs={"Metadata": {"k": "v"}},
        )

        f = BucketFile(exo.storage, {}, path="a", bucket=bucket)
        assert f.size == len(f_a.read_text())
        assert type(f.last_modification_date) == datetime
        assert f.metadata == {"k": "v"}
        assert f.content.read() == b"a"
        file_url = f.url
        assert file_url.startswith("https://sos-") and file_url.endswith("/a")

        # We retrieve the current owner of the file to maintain it in the updated ACP,
        # otherwise we'll lock ourselves out and we won't be able to delete the file
        # afterwards
        owner = exo.storage.boto.get_object_acl(Bucket=bucket.name, Key="a")["Owner"][
            "DisplayName"
        ]

        exo.storage.boto.put_object_acl(
            Bucket=bucket.name,
            Key="a",
            AccessControlPolicy={
                "Grants": [
                    {
                        "Grantee": {
                            "DisplayName": owner,
                            "ID": owner,
                            "Type": "CanonicalUser",
                        },
                        "Permission": "FULL_CONTROL",
                    },
                    {
                        "Grantee": {"Type": "Group", "URI": ACL_ALL_USERS},
                        "Permission": "READ",
                    },
                    {
                        "Grantee": {"Type": "Group", "URI": ACL_AUTHENTICATED_USERS},
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "DisplayName": "bob@example.net",
                            "ID": "bob@example.net",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "READ_ACP",
                    },
                    {
                        "Grantee": {
                            "DisplayName": "carl@example.net",
                            "ID": "carl@example.net",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "WRITE_ACP",
                    },
                ]
            },
        )

        # Retrieving the bucket ACL policy immediately after having set it can trigger
        # a race condition at SOS level due to eventual consistency data store, so as
        # a workaround we wait for a bit before getting our ACL policy back.
        sleep(3)

        acp = f.acl
        assert acp.full_control == owner
        assert acp.read == "ALL_USERS"
        assert acp.write == "AUTHENTICATED_USERS"
        assert acp.read_acp == "bob@example.net"
        assert acp.write_acp == "carl@example.net"

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="a")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale import *

_ALICE_API_KEY = "alice_api_key"
_ALICE_API_SECRET = "alice_api_secret"
_BOB_API_KEY = "bob_api_key"
_BOB_API_SECRET = "bob_api_secret"


class TestConfig:
    def test_params(self):
        exo = Exoscale(api_key=_ALICE_API_KEY, api_secret=_ALICE_API_SECRET)
        assert exo.api_key == _ALICE_API_KEY
        assert exo.api_secret == _ALICE_API_SECRET

    def test_envvars(self, monkeypatch):
        monkeypatch.setenv("EXOSCALE_API_KEY", _ALICE_API_KEY)
        monkeypatch.setenv("EXOSCALE_API_SECRET", _ALICE_API_SECRET)

        exo = Exoscale()
        assert exo.api_key == _ALICE_API_KEY
        assert exo.api_secret == _ALICE_API_SECRET

    def test_file_no_default_profile(self, tmp_path):
        f = tmp_path / "config.toml"

        f.write_text(
            """
[[profiles]]
name = "alice"
api_key = "{alice_api_key}"
api_secret = "{alice_api_secret}"

[[profiles]]
name = "bob"
api_key = "{bob_api_key}"
api_secret = "{bob_api_secret}"
""".format(
                alice_api_key=_ALICE_API_KEY,
                alice_api_secret=_ALICE_API_SECRET,
                bob_api_key=_BOB_API_KEY,
                bob_api_secret=_BOB_API_SECRET,
            )
        )

        exo = Exoscale(config_file=f)
        assert exo.api_key == _ALICE_API_KEY
        assert exo.api_secret == _ALICE_API_SECRET

    def test_file_default_profile(self, tmp_path):
        f = tmp_path / "config.toml"

        f.write_text(
            """
default_profile = "bob"

[[profiles]]
name = "alice"
api_key = "{alice_api_key}"
api_secret = "{alice_api_secret}"

[[profiles]]
name = "bob"
api_key = "{bob_api_key}"
api_secret = "{bob_api_secret}"
""".format(
                alice_api_key=_ALICE_API_KEY,
                alice_api_secret=_ALICE_API_SECRET,
                bob_api_key=_BOB_API_KEY,
                bob_api_secret=_BOB_API_SECRET,
            )
        )

        exo = Exoscale(config_file=f)
        assert exo.api_key == _BOB_API_KEY
        assert exo.api_secret == _BOB_API_SECRET

    def test_file_profile_param(self, tmp_path):
        f = tmp_path / "config.toml"

        f.write_text(
            """
[[profiles]]
name = "alice"
api_key = "{alice_api_key}"
api_secret = "{alice_api_secret}"

[[profiles]]
name = "bob"
api_key = "{bob_api_key}"
api_secret = "{bob_api_secret}"
""".format(
                alice_api_key=_ALICE_API_KEY,
                alice_api_secret=_ALICE_API_SECRET,
                bob_api_key=_BOB_API_KEY,
                bob_api_secret=_BOB_API_SECRET,
            )
        )

        exo = Exoscale(config_file=f, profile="bob")
        assert exo.api_key == _BOB_API_KEY
        assert exo.api_secret == _BOB_API_SECRET

        with pytest.raises(ConfigurationError) as excinfo:
            exo = Exoscale(config_file=f, profile="lolnope")
            assert exo is None
        assert excinfo.type == ConfigurationError
        assert 'profile "lolnope" not found' in str(excinfo.value)

    def test_file_errors(self, tmp_path):
        f = tmp_path / "config.toml"

        f.write_text("")
        with pytest.raises(ConfigurationError) as excinfo:
            exo = Exoscale(config_file=f)
            assert exo is None
        assert excinfo.type == ConfigurationError
        assert "no profiles configured" in str(excinfo.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeSecurityGroup:
    def test_add_rule(self, exo, sg, test_description):
        security_group = SecurityGroup._from_cs(exo.compute, sg())
        test_network_cidr = "1.1.1.1/32"
        test_port_ingress = "80-81"
        test_protocol_ingress = "tcp"
        test_port_egress = "53"
        test_protocol_egress = "udp"

        security_group_default = SecurityGroup._from_cs(
            exo.compute,
            exo.compute.cs.listSecurityGroups(
                securitygroupname="default", fetch_list=True
            )[0],
        )

        for rule in [
            SecurityGroupRule.ingress(
                description=test_description,
                security_group=security_group_default,
                port=test_port_ingress,
                protocol=test_protocol_ingress,
            ),
            SecurityGroupRule.egress(
                description=test_description,
                network_cidr=test_network_cidr,
                port=test_port_egress,
                protocol=test_protocol_egress,
            ),
        ]:
            security_group.add_rule(rule)

        res = exo.compute.cs.listSecurityGroups(id=security_group.id, fetch_list=True)
        assert len(res[0]["ingressrule"]) == 1
        assert res[0]["ingressrule"][0]["description"] == test_description
        assert res[0]["ingressrule"][0]["securitygroupname"] == "default"
        assert res[0]["ingressrule"][0]["protocol"] == test_protocol_ingress
        assert res[0]["ingressrule"][0]["startport"] == int(
            test_port_ingress.split("-")[0]
        )
        assert res[0]["ingressrule"][0]["endport"] == int(
            test_port_ingress.split("-")[1]
        )
        assert len(res[0]["egressrule"]) == 1
        assert res[0]["egressrule"][0]["description"] == test_description
        assert res[0]["egressrule"][0]["cidr"] == test_network_cidr
        assert res[0]["egressrule"][0]["protocol"] == test_protocol_egress
        assert str(res[0]["egressrule"][0]["startport"]) == test_port_egress
        assert str(res[0]["egressrule"][0]["endport"]) == test_port_egress

    def test_delete(self, exo, sg):
        security_group = SecurityGroup._from_cs(exo.compute, sg(teardown=False))
        security_group_name = security_group.name

        security_group.delete()
        assert security_group.id is None

        res = exo.compute.cs.listSecurityGroups(
            securitygroupname=security_group_name, fetch_list=True
        )
        assert len(res) == 0

    def test_properties(self, exo, sg):
        security_group = SecurityGroup._from_cs(exo.compute, sg())
        security_group_default = SecurityGroup._from_cs(
            exo.compute,
            exo.compute.cs.listSecurityGroups(
                securitygroupname="default", fetch_list=True
            )[0],
        )

        exo.compute.cs.authorizeSecurityGroupIngress(
            securitygroupid=security_group.id,
            usersecuritygrouplist={"group": "default"},
            protocol="tcp",
            startport=8000,
            endport=9000,
        )
        exo.compute.cs.authorizeSecurityGroupEgress(
            securitygroupid=security_group.id,
            protocol="udp",
            cidrlist="0.0.0.0/0",
            startport=53,
            endport=53,
        )

        security_group_ingress_rules = list(security_group.ingress_rules)
        assert len(security_group_ingress_rules) == 1
        assert security_group_ingress_rules[0].type == "ingress"
        assert (
            security_group_ingress_rules[0].security_group.id
            == security_group_default.id
        )
        assert security_group_ingress_rules[0].port == "8000-9000"
        assert security_group_ingress_rules[0].protocol == "tcp"

        security_group_egress_rules = list(security_group.egress_rules)
        assert len(security_group_egress_rules) == 1
        assert security_group_egress_rules[0].type == "egress"
        assert security_group_egress_rules[0].network_cidr == "0.0.0.0/0"
        assert security_group_egress_rules[0].port == "53"
        assert security_group_egress_rules[0].protocol == "udp"
